/* 
 * plugins.c
 *
 * hashkill - a hash cracking tool
 * Copyright (C) 2010 Milen Rangelov <gat3way@gat3way.eu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <dlfcn.h>
#include <sys/types.h>
#include <errno.h>
#include <string.h>
#include "plugins.h"
#include "err.h"
#include "hashinterface.h"
#include "plugins/plugin.h"


/* Global variables */
static char currentplugin[HASHFILE_MAX_LINE_LENGTH];	// current plugin in use


/* Function prototypes */
hash_stat print_plugins_summary(char *plugindir);
void print_plugin_detailed(char *plugin);
char *get_current_plugin(void);
void set_current_plugin(const char *plugin);
hash_stat load_plugin(void);





/* Print info on all available plugins */
hash_stat print_plugins_summary(char *plugindir)
{
    void *dlhandle;
    char * (*hash_plugin_summary)(void);
    struct dirent **dentrylist;
    char soname[1024];
    DIR *dir;
    int count=-1,i=0;

    dir=opendir(plugindir);
    if (!dir)
    {
	elog("Cannot open plugins dir: %s", plugindir);
	return hash_err;
    }
    closedir(dir);

    
    hlog("Plugins list: %s\n\n","");
    
    count = scandir(plugindir, &dentrylist, 0, alphasort);
    do
    {
	errno = 0;
	if (strstr(dentrylist[i]->d_name, ".so"))
	{
	    snprintf(soname,1024,"%s/%s", plugindir, dentrylist[i]->d_name);
	    dlhandle=dlopen(soname,RTLD_LAZY);
	    if (dlhandle)
	    {
	        *(void **) (&hash_plugin_summary) = dlsym(dlhandle, "hash_plugin_summary");
	        if (hash_plugin_summary != NULL) printf("%s\n",hash_plugin_summary());
	        dlclose(dlhandle);
	    }
	}
	i++;
    } while (i<count);
    free(dentrylist);
    printf("\n");
    return hash_ok;
}


/* Print detailed info on plugin */
void print_plugin_detailed(char *plugin)
{
    char soname[1024];
    void *dlhandle;
    char * (*hash_plugin_detailed)(void);

    snprintf(soname,1024,"/usr/share/hashkill/plugins/%s.so", plugin);
    dlhandle=dlopen(soname,RTLD_LAZY);
    if (dlhandle)
    {
        *(void **) (&hash_plugin_detailed) = dlsym(dlhandle, "hash_plugin_detailed");
        if (hash_plugin_detailed != NULL) printf("\n%s\n\n",hash_plugin_detailed());
        dlclose(dlhandle);
    }
    else 
    {
	elog("Cannot open plugin library: %s\n",soname);
    }
}



/* Get current plugin */
char *get_current_plugin(void)
{
    return (char *)&currentplugin;
}


/* Set current plugin */
void set_current_plugin(const char *plugin)
{
    strcpy(currentplugin,plugin);
}


/* Load plugin */
hash_stat load_plugin(void)
{
    char soname[1024];
    void *dlhandle;

    snprintf(soname,1024,"/usr/share/hashkill/plugins/%s.so", get_current_plugin());
    dlhandle=dlopen(soname,RTLD_LAZY);
    if (dlhandle)
    {
        *(hash_stat **) (&hash_plugin_parse_hash) = dlsym(dlhandle, "hash_plugin_parse_hash");
        *(hash_stat **) (&hash_plugin_check_hash) = dlsym(dlhandle, "hash_plugin_check_hash");
        *(int **) (&hash_plugin_hash_length) = dlsym(dlhandle, "hash_plugin_hash_length");
        *(int **) (&hash_plugin_is_raw) = dlsym(dlhandle, "hash_plugin_is_raw");
        *(int **) (&hash_plugin_is_special) = dlsym(dlhandle, "hash_plugin_is_special");
        *(void **) (&get_vector_size) = dlsym(dlhandle, "get_vector_size");
	
	if ( (!hash_plugin_parse_hash) || (!hash_plugin_check_hash) || (!hash_plugin_hash_length))
	{
	    elog("Plugin %s does not export all the necessary functions!\n", get_current_plugin());
	    return hash_err;
	}
	
// import register functions
	void(*register_add_username)(void (*add_username)(const char *username)) = dlsym(dlhandle, "register_add_username");
	void(*register_add_hash)(void (*add_hash)(const char *hash, int len)) = dlsym(dlhandle, "register_add_hash");
	void(*register_add_salt)(void (*add_salt)(const char *salt)) = dlsym(dlhandle, "register_add_salt");
	void(*register_add_salt2)(void (*add_salt2)(const char *salt2)) = dlsym(dlhandle, "register_add_salt2");
	void(*register_md5)(void (*md5)(char *plaintext[VECTORSIZE], char *hash[VECTORSIZE], int len, int threadid)) = dlsym(dlhandle, "register_md5");
	void(*register_md5_unicode)(void (*md5_unicode)(char *plaintext[VECTORSIZE], char *hash[VECTORSIZE], int len[VECTORSIZE])) = dlsym(dlhandle, "register_md5_unicode");
	void(*register_md5_unicode_slow)(void (*md5_unicode_slow)(char *plaintext[VECTORSIZE], char *hash[VECTORSIZE], int len[VECTORSIZE])) = dlsym(dlhandle, "register_md5_unicode_slow");
	void(*register_md5_slow)(void (*md5_slow)(char *plaintext[VECTORSIZE], char *hash[VECTORSIZE], int len, int threadid)) = dlsym(dlhandle, "register_md5_slow");
	void(*register_md4)(void (*md4)(char *plaintext[VECTORSIZE], char *hash[VECTORSIZE], int len[VECTORSIZE])) = dlsym(dlhandle, "register_md4");
	void(*register_md5_hex)(void (*md5_hex)(const char *hash[VECTORSIZE], char *hashhex[VECTORSIZE])) = dlsym(dlhandle, "register_md5_hex");
	void(*register_sha1)(void (*md5)(char *plaintext[VECTORSIZE], char *hash[VECTORSIZE], int len, int threadid)) = dlsym(dlhandle, "register_sha1");
	void(*register_sha1_unicode)(void (*sha1_unicode)(char *plaintext[VECTORSIZE], char *hash[VECTORSIZE], int len[VECTORSIZE])) = dlsym(dlhandle, "register_sha1_unicode");
	void(*register_sha1_slow)(void (*sha1_slow)(char *plaintext[VECTORSIZE], char *hash[VECTORSIZE], int len[VECTORSIZE])) = dlsym(dlhandle, "register_sha1_slow");
	void(*register_sha1_hex)(void (*md5_hex)(const char *hash[VECTORSIZE], char *hashhex[VECTORSIZE])) = dlsym(dlhandle, "register_sha1_hex");
	void(*register_sha256_unicode)(void (*sha256_unicode)(const char *plaintext[VECTORSIZE], char *hash[VECTORSIZE], int len[VECTORSIZE])) = dlsym(dlhandle, "register_sha256_unicode");
	void(*register_sha256_hex)(void (*sha256_hex)(const char *hash[VECTORSIZE], char *hashhex[VECTORSIZE])) = dlsym(dlhandle, "register_sha256_hex");
	void(*register_sha512_unicode)(void (*sha512_unicode)(const char *plaintext[VECTORSIZE], char *hash[VECTORSIZE], int len[VECTORSIZE])) = dlsym(dlhandle, "register_sha512_unicode");
	void(*register_sha512_hex)(void (*sha512_hex)(const char *hash[VECTORSIZE], char *hashhex[VECTORSIZE])) = dlsym(dlhandle, "register_sha512_hex");
	void(*register_fcrypt)(void (*fcrypt)(const char *password[VECTORSIZE], const char *salt, char *ret[VECTORSIZE])) = dlsym(dlhandle, "register_fcrypt");
	void(*register_PEM_readfile)(void (*PEM_readfile)(const char *passphrase, int *RSAret)) = dlsym(dlhandle, "register_PEM_readfile");
	void(*register_new_biomem)(void (*new_biomem)(FILE *file)) = dlsym(dlhandle, "register_new_biomem");
	void(*register_pbkdf2)(void (*pbkdf2)(const char *pass, unsigned char *salt, int saltlen, int iter, int keylen, unsigned char *out)) = dlsym(dlhandle, "register_pbkdf2");
	void(*register_hmac_sha1_file)(void (*hmac_sha1_file)(void *key, int keylen, char *filename, long offset, long size, unsigned char *output, int outputlen)) = dlsym(dlhandle, "register_hmac_sha1_file");
	void(*register_ripemd160)(void (*ripemd160)(const char *plaintext[VECTORSIZE], char *hash[VECTORSIZE], int len[VECTORSIZE])) = dlsym(dlhandle, "register_ripemd160");
	void(*register_pbkdf512)(void (*pbkdf512)(const char *pass, unsigned char *salt, int saltlen, int iter, int keylen, unsigned char *out)) = dlsym(dlhandle, "register_pbkdf512");
	void(*register_aes_encrypt)(void (*aes_encrypt)(const unsigned char *key, int keysize, const unsigned char *in, int len, unsigned char *vec, unsigned char *out, int mode)) = dlsym(dlhandle, "register_aes_encrypt");
	void(*register_aes_decrypt)(void (*aes_decrypt)(const unsigned char *key, int keysize, const unsigned char *in, int len, unsigned char *vec, unsigned char *out, int mode)) = dlsym(dlhandle, "register_aes_decrypt");
	void(*register_des_ecb_encrypt)(void (*des_ecb_encrypt)(const unsigned char *key, int keysize, const unsigned char *in[VECTORSIZE], int len, unsigned char *out[VECTORSIZE], int mode)) = dlsym(dlhandle, "register_des_ecb_encrypt");
	void(*register_des_ecb_decrypt)(void (*des_ecbdecrypt)(const unsigned char *key, int keysize, const unsigned char *in, int len, unsigned char *out, int mode)) = dlsym(dlhandle, "register_des_ecb_decrypt");
	void(*register_des_cbc_encrypt)(void (*des_ecb_encrypt)(const unsigned char *key[VECTORSIZE], int keysize, const unsigned char *in[VECTORSIZE], int len[VECTORSIZE], unsigned char *out[VECTORSIZE], unsigned char *iv[VECTORSIZE], int mode)) = dlsym(dlhandle, "register_des_cbc_encrypt");
	void(*register_lm)(void (*lm)(const unsigned char *in[VECTORSIZE], unsigned char *out[VECTORSIZE])) = dlsym(dlhandle, "register_lm");



        if ((hash_plugin_parse_hash != NULL) && (hash_plugin_check_hash != NULL) && (hash_plugin_is_raw != NULL) && (hash_plugin_is_special != NULL))
        {
// register our callbacks
	    register_add_username(hash_proto_add_username);
	    register_add_hash(hash_proto_add_hash);
	    register_add_salt(hash_proto_add_salt);
	    register_add_salt2(hash_proto_add_salt2);
	    register_md5(hash_proto_md5);
	    register_md5_unicode(hash_proto_md5_unicode);
	    register_md5_unicode_slow(hash_proto_md5_unicode_slow);
	    register_md5_slow(hash_proto_md5_slow);
	    register_md4(hash_proto_md4);
	    register_md5_hex(hash_proto_md5_hex);
	    register_sha1(hash_proto_sha1);
	    register_sha1_unicode(hash_proto_sha1_unicode);
	    register_sha1_slow(hash_proto_sha1_slow);
	    register_sha1_hex(hash_proto_sha1_hex);
	    register_sha256_unicode(hash_proto_sha256_unicode);
	    register_sha256_hex(hash_proto_sha256_hex);
	    register_sha512_unicode(hash_proto_sha512_unicode);
	    register_sha512_hex(hash_proto_sha512_hex);
	    register_fcrypt(hash_proto_fcrypt);
	    register_new_biomem(hash_proto_new_biomem);
	    register_PEM_readfile(hash_proto_PEM_readfile);
	    register_pbkdf2(hash_proto_pbkdf2);
	    register_hmac_sha1_file(hash_proto_hmac_sha1_file);
	    register_ripemd160(hash_proto_ripemd160);
	    register_pbkdf512(hash_proto_pbkdf512);
	    register_aes_encrypt(hash_proto_aes_encrypt);
	    register_aes_decrypt(hash_proto_aes_decrypt);
	    register_des_ecb_encrypt(hash_proto_des_ecb_encrypt);
	    register_des_ecb_decrypt(hash_proto_des_ecb_decrypt);
	    register_des_cbc_encrypt(hash_proto_des_cbc_encrypt);
	    register_lm(hash_proto_lm);
	    
	    /* set vector size hackery */
	    if ( (strcmp(get_current_plugin(),"desunix")==0) || (strcmp(get_current_plugin(),"lm")==0) || (strcmp(get_current_plugin(),"oracle-old")==0)) {vectorsize = 128; get_vector_size(128);}
	    else if ( (strcmp(get_current_plugin(),"privkey")==0)  || (strcmp(get_current_plugin(),"zip")==0) )
	    {vectorsize = 1; get_vector_size(1);}
	    else {vectorsize = 16; get_vector_size(16);}
	    
	    hash_is_raw = hash_plugin_is_raw();
	    hlog("Plugin \'%s\' loaded successfully\n",get_current_plugin());
	}
	else 
	{
	    elog("Plugin \'%s\' could not be loaded\n",get_current_plugin());
	    dlclose(dlhandle);
	    return hash_err;
	}
    }
    else 
    {
	elog("Cannot open plugin library: %s\n",soname);
	return hash_err;
    }
    return hash_ok;
}

