/* 
 * threads.c
 *
 * hashkill - a hash cracking tool
 * Copyright (C) 2010 Milen Rangelov <gat3way@gat3way.eu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


//#define DEBUG

#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/sysinfo.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <stdint.h>
#include <math.h>
#include <alloca.h>
#include <termios.h>
#include <openssl/bn.h>
#include <openssl/crypto.h>
#include "err.h"
#include "threads.h"
#include "hashinterface.h"
#include "plugins.h"
#include "sessions.h"


/* Global static variables */
static int attack_over;					// flag indicating attack is done
static int hash_len; 					// strlen(hash)
static uint64_t attack_overall_count;			// max. (attack metrics)
static uint64_t attack_current_count;			// cur. (attack metrics)
static uint64_t attack_current_length;			// bruteforce, current string len
static int *worker_thread_keys[HASHKILL_MAXTHREADS];	// used to set thread args
static char *attack_current_str;			// current bruteforce string
static uint64_t bruteforce_current_position;		// set by sessions.c
static int bruteforce_current_length;			// set by sessions.c
static long dictionary_current_pos;			// set by sessions.c
static int qflag;
static uint64_t markov_current_pos;
static uint64_t markov_calculated_max_pos;
static int markov_current_len;
static char *fillbuf[VECTORSIZE+1];
static char bitmap1[256][256];
static char bitmap2[256][256];
static char bitmap3[256][256];
static char bitmap4[256][256];

/* OpenSSL thread-safety variables */
static pthread_mutex_t *lock_cs;			// mutexes
static long            *lock_count;			// mutex_count


/* queue-ready static vars */
static int pushready[HASHKILL_MAXTHREADS][32];
static int popready[HASHKILL_MAXTHREADS][32];
/* prefix len stored to speed up cracking */
static int bruteforce_prefixlen;
/* Used to get attack timing */
static time_t time1,time2;
/* We use this to restore a bruteforce session */
static char saved_bf_string[16];



/* Function prototypes */
static void * start_monitor_thread(void *arg);
static void * start_monitor_info_thread(void *arg);

static void SSL_thread_cleanup( void );
static void SSL_pthreads_locking_callback(int mode, int type, char *file, int line);
static unsigned long SSL_pthreads_thread_id(void);
static void SSL_thread_setup(void);
static void pushqueue(char *ptext[VECTORSIZE], int queueid);
static void *popqueue(int queueid, int *lens);
static void thread_wait_empty_queues(void);
static void thread_attack_worker_nonlinear(int *who);
static void thread_attack_worker_linear(int *who);
static void * worker_thread_action(void *arg);
hash_stat spawn_threads(unsigned int num, unsigned int queue_size);
unsigned int hash_num_cpu(void);
unsigned long hash_get_total_memory(void);
static inline void bruteforce_fillqueue(char *plaintext);
static void bruteforce_permute(char *str, int n, const char *baseset);
hash_stat main_thread_hybrid(void);
hash_stat main_thread_dictionary(void);
hash_stat main_thread_bruteforce(void);
void thread_attack_cleanup(void);
void thread_set_bf_len(int len);
void thread_set_bf_str(char *str);
void thread_set_dict_pos(long pos);
void thread_set_bf_pos(uint64_t pos);


#if (defined __i386) || (defined __i486) ||(defined __i586) || (defined __i686) || (defined __x86_64)

#define strlen my_strlen
/* Optimized strlen() for x86 architectures */
inline size_t my_strlen(const char *s) {
    size_t len = 0;
    for(;;) {
        unsigned x = *(unsigned*)s;
        if((x & 0xFF) == 0) return len;
        if((x & 0xFF00) == 0) return len + 1;
        if((x & 0xFF0000) == 0) return len + 2;
        if((x & 0xFF000000) == 0) return len + 3;
        s += 4, len += 4;
    }
}


#define memcmp memcmp_word_asm
/* Optimized memcmp() for x86 architectures */
static inline unsigned int bswap(unsigned int x)
{
    asm("bswap %0" : "=r" (x) : "0" (x)); return x;
}

inline int memcmp_word_asm(const void *a, const void *b, size_t n)
{
    int num_ints = n / sizeof(int);
    int extra_bytes = n % sizeof(int);
    unsigned int *ua = (unsigned int *)a;
    unsigned int *ub = (unsigned int *)b;
    unsigned int *ua_max = ua + num_ints;
    unsigned char *uac;
    unsigned char *ubc;
    unsigned char *uac_max;

    while (ua < ua_max)
    {
	if (*ua == *ub)
	{
	    ua++;
	    ub++;
	} 
	else if (bswap(*ua) < bswap(*ub)) return -1;
	else return 1;
    }
    uac = (unsigned char *)ua;
    ubc = (unsigned char *)ub;
    uac_max = uac + extra_bytes;
    while (uac < uac_max)
    {
	if (*uac == *ubc)
	{
	    uac++;
	    ubc++;
	}
	else if (*uac < *ubc) return -1;
	else return 1;
    }
    return 0;
}

#endif


int fastcompare(const char *s1, const char *s2, int length)
{
    unsigned int i;

    for(i=0; i < length; i++)
    {
        if(*(s1+i) != *(s2+i)) return 1;
    }
    return 0;
}




/* Cleanup openssl thread-safe callbacks */
static void SSL_thread_cleanup(void) 
{
    int x;
    CRYPTO_set_locking_callback(NULL);
    for( x = 0; x < CRYPTO_num_locks(); x++ ) 
    {
	pthread_mutex_destroy(&(lock_cs[x]));
    }
    OPENSSL_free(lock_cs);
    OPENSSL_free(lock_count);
}


/* The openssl crypto locking callback */
static void SSL_pthreads_locking_callback(int mode, int type, char *file, int line) 
{
    if(mode & CRYPTO_LOCK)
    {
	pthread_mutex_lock(&(lock_cs[type]));
	lock_count[type]++;
    } 
    else
    { 
	pthread_mutex_unlock(&(lock_cs[type]));
    }
}


/* The next callback needed */
static unsigned long SSL_pthreads_thread_id(void) 
{
  unsigned long ret;
  ret = (unsigned long)pthread_self();

  return(ret);
}



/* setup open crypto callbacks */
static void SSL_thread_setup(void) 
{
    int x;

#define OPENSSL_THREAD_DEFINES
#include <openssl/opensslconf.h>
#if defined(THREADS) || defined(OPENSSL_THREADS)
#else
    fprintf(stderr, "WARNING: your openssl libraries were compiled without thread support\n");
    pthread_sleep_np( 2 );
#endif
    lock_cs    = (pthread_mutex_t*)OPENSSL_malloc(CRYPTO_num_locks()*sizeof(pthread_mutex_t));
    lock_count = (long*)OPENSSL_malloc(CRYPTO_num_locks() * sizeof(long));
    for( x = 0; x < CRYPTO_num_locks(); x++ )
    {
	lock_count[x] = 0;
	pthread_mutex_init(&(lock_cs[x]), NULL);
    }
    CRYPTO_set_id_callback((unsigned long (*)())SSL_pthreads_thread_id);
    CRYPTO_set_locking_callback((void (*)())SSL_pthreads_locking_callback);
}




/* Push plaintext into queue */
static inline void pushqueue(char *ptext[VECTORSIZE], int queueid)
{
    int val;
    int a;
    
    
    val = hash_thread_queue_counts[queueid][0];
    if (val == 0)
    {
	pthread_mutex_lock(&queuelock[queueid]);
	pushready[queueid][0] = 1;
    }
    
    for (a=0;a<vectorsize;a++) {
	memcpy(hash_queue[queueid][val].plaintext[a], ptext[a], HASHFILE_MAX_PLAIN_LENGTH);
	//hash_queue[queueid][val].plaintext[a][HASHFILE_MAX_PLAIN_LENGTH] = 0;
	hash_queue[queueid][val].lens[a] = strlen(ptext[a]);
    }
    
    val++;
    hash_thread_queue_counts[queueid][0]++;

    if (val == hash_queue_size-1)
    {
	pushready[queueid][0] = 0;
	pthread_mutex_unlock(&queuelock[queueid]);
    }
}


/* 
    Pop plaintext from queue (flag indicates locking)
    queue mutex MUST be locked prior to calling if flag=1
*/
static inline void *popqueue(int queueid, int *lens)
{
    int val = 0;
    int a;
    char *str;

    again:
    val = hash_thread_queue_counts[queueid][0];
    if (val == 0) 
    {
	if (popready[queueid][0] == 1)
	{
	    popready[queueid][0] = 0;
	    pthread_mutex_unlock(&queuelock[queueid]);
	}
	else 
	{
	    usleep(1000);
	    goto again;
	}
    }
    else
    {
	if (popready[queueid][0] == 0)
	{
	    pthread_mutex_lock(&queuelock[queueid]);
	    popready[queueid][0] = 1;
	}
	
	*lens = val;
	hash_thread_queue_counts[queueid][0]--;
	return (void *)&hash_queue[queueid][val].plaintext;
    }
    return (void *)&hash_queue[queueid][val].plaintext;

}




/* Wait until queues have been emptied */
static void thread_wait_empty_queues(void)
{
    int cnt,a;
    unsigned int flag = 1;

    if (attack_over == 2) return;
    while (flag == 1)
    {
	flag = 0;
	for (cnt=0;cnt < hash_num_threads;cnt++) 
	{
	    usleep(1000);
	    a = hash_thread_queue_counts[cnt][0];
	    if (a > 0) flag = 1;
	}
    }
    attack_over = 2;
    for (cnt=0;cnt < hash_num_threads;cnt++) pthread_cancel(hash_threads[cnt]);
    SSL_thread_cleanup();
}



/* worker thread - nonlinear complexity */
static void thread_attack_worker_nonlinear(int *who)
{
    struct  hash_list_s  *mylist;
    struct hash_list_s * mylist1;
    struct hash_list_s * addlist;
    int flag;
    int whoami;
    int val=0;
    int num;
    char *myhash[VECTORSIZE];
    char *plain[VECTORSIZE];
    int zero;
    int res;
    int a;
    
    for (flag=0;flag<=vectorsize;flag++) myhash[flag] = alloca(HASHFILE_MAX_LINE_LENGTH+1);
    for (flag=0;flag<=vectorsize;flag++) plain[flag] = alloca(HASHFILE_MAX_LINE_LENGTH+1);

    whoami = (int )*who;
    while (1)
    {
	//if (attack_over == 2) pthread_exit(NULL);
	memcpy(plain,popqueue(whoami,&zero),vectorsize*sizeof(char *));
	hash_thread_processed[whoami][0]+=vectorsize;
	mylist = hash_list;
	while (mylist)
	{
	    num = zero;
	    res = hash_plugin_check_hash(mylist->hash, (const char **)&plain,mylist->salt, myhash, mylist->username, &num, whoami);
	    flag = 0;
	    if (res==hash_ok)
	    {
		pthread_mutex_lock(&crackedmutex);
		addlist = cracked_list;
		#ifdef DEBUG
		hlog("Cracked hash: %s plaintext: %s thread: %d\n",mylist->hash, plain[num], whoami);
		#endif
		while (addlist)
		{
		    if ( (strcmp(addlist->username, mylist->username) == 0) && (strcmp(addlist->hash, mylist->hash) == 0)
		    && (strcmp(addlist->salt, mylist->salt) == 0))
		    flag = 1;
		    addlist = addlist->next;
		}
		pthread_mutex_unlock(&crackedmutex);
		if (flag == 0) add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain[num]);
	    }
	    if (mylist) mylist = mylist->next;
	}
    }
    pthread_exit(NULL);
}




/* worker thread - nonlinear complexity - bruteforce */
static void thread_attack_worker_nonlinear_bruteforce(int *who)
{
    struct  hash_list_s  *mylist;
    struct hash_list_s * mylist1;
    struct hash_list_s * addlist;
    int flag;
    int whoami;
    int val=0;
    int num;
    char *myhash[VECTORSIZE];
    char *plain[VECTORSIZE];
    int zero,d,e,lens,res;
    
    for (flag=0;flag<=vectorsize;flag++) myhash[flag] = alloca(HASHFILE_MAX_LINE_LENGTH+1);
    for (flag=0;flag<=vectorsize;flag++) plain[flag] = alloca(HASHFILE_MAX_LINE_LENGTH+1);

    whoami = (int )*who;
    while (1)
    {
	//if (attack_over == 2) pthread_exit(NULL);
	memcpy(plain,popqueue(whoami,&lens),vectorsize*sizeof(char *));
	for (e=0;e<strlen(bruteforce_charset);e++)
	{
	    for (d=0;d<vectorsize;d++) 
	    {
		plain[d][hash_queue[whoami][lens].lens[d]] = bruteforce_charset[e];
		plain[d][hash_queue[whoami][lens].lens[d]+1] = 0;
		bzero(plain[d]+hash_queue[whoami][lens].lens[d]+1,10);
		hash_queue[whoami][lens].lens[d]++;
	    }
	
	    hash_thread_processed[whoami][0]+=vectorsize;
	    mylist = hash_list;
	    while (mylist)
	    {
		num = lens;
		res = hash_plugin_check_hash(mylist->hash, (const char **)&plain,mylist->salt, myhash, mylist->username, &num, whoami);
		#ifdef DEBUG
		hlog("Cracked hash: %s plaintext: %s thread: %d\n",mylist->hash, plain, whoami);
		#endif
		if (res==hash_ok)
		{
		    flag = 0;
		    pthread_mutex_lock(&crackedmutex);
		    addlist = cracked_list;
		    while (addlist)
		    {
		        if ( (strcmp(addlist->username, mylist->username) == 0) && (strcmp(addlist->hash, mylist->hash) == 0)
				&& (strcmp(addlist->salt, mylist->salt) == 0))
				    flag = 1;
			addlist = addlist->next;
		    }
		    pthread_mutex_unlock(&crackedmutex);
		    if (flag == 0) add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain[num]);
		}
	    	if (mylist) mylist = mylist->next;
	    }
	    for (d=0;d<vectorsize;d++) 
	    {
		hash_queue[whoami][lens].lens[d]--;
	    }
	}
    }
    pthread_exit(NULL);
}



/* worker thread - nonlinear complexity - markov */
static void thread_attack_worker_nonlinear_markov(int *who)
{
    struct  hash_list_s  *mylist;
    struct hash_list_s * mylist1;
    struct hash_list_s * addlist;
    int flag;
    int whoami;
    int val=0;
    int num;
    char *myhash[VECTORSIZE];
    char *plain[VECTORSIZE];
    int zero,d,e,lens;
    
    for (flag=0;flag<=vectorsize;flag++) myhash[flag] = alloca(HASHFILE_MAX_LINE_LENGTH+1);
    for (flag=0;flag<=vectorsize;flag++) plain[flag] = alloca(HASHFILE_MAX_LINE_LENGTH+1);

    whoami = (int )*who;
    while (1)
    {
	//if (attack_over == 2) pthread_exit(NULL);
	memcpy(plain,popqueue(whoami,&lens),vectorsize*sizeof(char *));
	for (e=0;e<strlen(markov_charset);e++)
	{
	    for (d=0;d<vectorsize;d++) 
	    {
		plain[d][hash_queue[whoami][lens].lens[d]] = markov_charset[e];
		plain[d][hash_queue[whoami][lens].lens[d]+1] = 0;
		bzero(plain[d]+hash_queue[whoami][lens].lens[d]+1,10);
		hash_queue[whoami][lens].lens[d]++;
	    }
	
	    hash_thread_processed[whoami][0] += vectorsize;
	    mylist = hash_list;
	    while (mylist)
	    {
		num = lens;
		while ( (hash_plugin_check_hash(mylist->hash, (const char **)&plain,
		    mylist->salt, myhash, mylist->username, &num, whoami) == hash_ok))
		{
			#ifdef DEBUG
			hlog("Cracked hash: %s plaintext: %s thread: %d\n",mylist->hash, plain, whoami);
			#endif
			flag = 0;
			pthread_mutex_lock(&crackedmutex);
			addlist = cracked_list;
			while (addlist)
			{
			    if ( (strcmp(addlist->username, mylist->username) == 0) && (strcmp(addlist->hash, mylist->hash) == 0)
				&& (strcmp(addlist->salt, mylist->salt) == 0))
				    flag = 1;
			    addlist = addlist->next;
			}
			pthread_mutex_unlock(&crackedmutex);
			if (flag == 0) add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain[num]);
			plain[num][0]=0;
			plain[num][1]=0;
			hash_queue[whoami][lens].lens[num]=2;
		}
	    	if (mylist) mylist = mylist->next;
	    }
	    for (d=0;d<vectorsize;d++) 
	    {
		hash_queue[whoami][lens].lens[d]--;
	    }
	}
    }
    pthread_exit(NULL);
}









/* worker thread - linear complexity */
static void thread_attack_worker_linear(int *who)
{
    struct  hash_list_s  *mylist, *addlist;
    int whoami;
    int num,a;
    unsigned int val=0;
    char *myhash[VECTORSIZE];
    char *plain[VECTORSIZE];
    int flag = 0;
    int zero;
    int d;
    int res=hash_err;

    for (flag=0;flag<=vectorsize;flag++) myhash[flag] = alloca(HASHFILE_MAX_LINE_LENGTH+1);
    whoami = (int )*who;
    while (1)
    {

	memcpy(plain,popqueue(whoami,&zero),vectorsize*sizeof(char *));
	again:
	mylist = hash_list;
	if (mylist)
	{
	    num = zero;
	    res=hash_plugin_check_hash(mylist->hash, (const char **)plain,
		    mylist->salt, myhash, mylist->username, &num, whoami);
	}
	for (d=0;d<vectorsize;d++) if (bitmap1[myhash[d][0]&255][myhash[d][1]&255]==1) 
	if  (bitmap2[myhash[d][2]&255][myhash[d][3]&255]==1) 
	if (bitmap3[myhash[d][4]&255][myhash[d][5]&255]==1)
	if (bitmap4[myhash[d][6]&255][myhash[d][7]&255]==1)
	while (mylist) 
	{
	    for (a=0;a<vectorsize;a++) 
	    {	
		if (memcmp(mylist->hash, myhash[a], hash_ret_len) == 0)
		{

			#ifdef DEBUG
			hlog("Cracked uname: %s plaintext: %s thread: %d\n",mylist->username, plain, whoami);
			#endif
			flag = 0;
			pthread_mutex_lock(&crackedmutex);
			addlist = cracked_list;
			while (addlist)
			{
			    if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
				&& (strcmp(addlist->salt, mylist->salt) == 0))
				    flag = 1;
			    addlist = addlist->next;
			}
			pthread_mutex_unlock(&crackedmutex);
			if (flag == 0) 
			{
			    add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain[a]);
			    plain[num][0]=0;
			    plain[num][1]=0;
			    hash_queue[whoami][zero].lens[num]=2;
			}
		}
	    }
	    if (mylist) mylist = mylist->next;
	}
	hash_thread_processed[whoami][0]+=vectorsize;
    }
}




/* worker thread - linear complexity - bruteforce*/
static void thread_attack_worker_linear_bruteforce(int *who)
{
    struct  hash_list_s  *mylist, *addlist;
    int whoami;
    int num,a;
    unsigned int val=0;
    char *myhash[VECTORSIZE];
    char *plain[VECTORSIZE];
    int flag = 0;
    void *ptr;
    int d,e;
    int lens;
    int res=hash_err;
    char username[128];

    for (flag=0;flag<=vectorsize;flag++) plain[flag] = alloca(HASHFILE_MAX_LINE_LENGTH+1);
    for (flag=0;flag<=vectorsize;flag++) myhash[flag] = alloca(HASHFILE_MAX_LINE_LENGTH+1);
    whoami = (int )*who;
    while (1)
    {
	memcpy(plain,popqueue(whoami,&lens),vectorsize*sizeof(char *));
	for (e=0;e<strlen(bruteforce_charset);e++)
	{
	    for (d=0;d<vectorsize;d++) 
	    {
		plain[d][hash_queue[whoami][lens].lens[d]] = markov_charset[e];
		plain[d][hash_queue[whoami][lens].lens[d]+1] = 0;
		bzero(plain[d]+hash_queue[whoami][lens].lens[d]+1,10);
		hash_queue[whoami][lens].lens[d]++;
	    }
	    
	    mylist = hash_list;
	    if (mylist)
	    {
		num = lens;
		res = hash_plugin_check_hash(mylist->hash, (const char **)plain,
			mylist->salt, myhash, mylist->username, &num, whoami);
	    }
	    
	    for (d=0;d<vectorsize;d++) if (bitmap1[myhash[d][0]&255][myhash[d][1]&255]==1) 
	    if (bitmap2[myhash[d][2]&255][myhash[d][3]&255]==1) 
	    if (bitmap3[myhash[d][4]&255][myhash[d][5]&255]==1)
	    if (bitmap4[myhash[d][6]&255][myhash[d][7]&255]==1)
	    while (mylist) 
	    {
		for (a=0;a<vectorsize;a++) 
		{	
		    if (memcmp(mylist->hash, myhash[a], hash_ret_len) == 0)
		    {

			#ifdef DEBUG
			hlog("Cracked uname: %s plaintext: %s thread: %d\n",mylist->username, plain, whoami);
			#endif
			flag = 0;
			pthread_mutex_lock(&crackedmutex);
			addlist = cracked_list;
			while (addlist)
			{
			    if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
				&& (strcmp(addlist->salt, mylist->salt) == 0))
				    flag = 1;
			    addlist = addlist->next;
			}
			pthread_mutex_unlock(&crackedmutex);
			if (flag == 0) 
			{
			    add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain[a]);
			}
		    }
		}
		if (mylist) mylist = mylist->next;
	    }
	    for (d=0;d<vectorsize;d++) 
	    {
		hash_queue[whoami][lens].lens[d]--;
	    }
	}
	hash_thread_processed[whoami][0]+=vectorsize*strlen(bruteforce_charset);
    }

}


/* worker thread - linear complexity - markov*/
static void thread_attack_worker_linear_markov(int *who)
{
    struct  hash_list_s  *mylist, *addlist;
    int whoami;
    int num,a;
    unsigned int val=0;
    char *myhash[VECTORSIZE];
    char *plain[VECTORSIZE];
    int flag = 0;
    void *ptr;
    int d,e;
    int lens;
    int res;

    for (flag=0;flag<=vectorsize;flag++) plain[flag] = alloca(HASHFILE_MAX_LINE_LENGTH+1);
    for (flag=0;flag<=vectorsize;flag++) myhash[flag] = alloca(HASHFILE_MAX_LINE_LENGTH+1);
    whoami = (int )*who;
    while (1)
    {
	memcpy(plain,popqueue(whoami,&lens),vectorsize*sizeof(char *));
	for (e=0;e<strlen(markov_charset);e++)
	{
	    for (d=0;d<vectorsize;d++) 
	    {
		plain[d][hash_queue[whoami][lens].lens[d]] = markov_charset[e];
		plain[d][hash_queue[whoami][lens].lens[d]+1] = 0;
		bzero(plain[d]+hash_queue[whoami][lens].lens[d]+1,10);
		hash_queue[whoami][lens].lens[d]++;
	    }
	    
	    mylist = hash_list;
	    if (mylist)
	    {
		num = lens;
		res = hash_plugin_check_hash(mylist->hash, (const char **)plain,
			mylist->salt, myhash, mylist->username, &num, whoami);
	    }
	    
	    for (d=0;d<vectorsize;d++) if (bitmap1[myhash[d][0]&255][myhash[d][1]&255]==1) 
	    if (bitmap2[myhash[d][2]&255][myhash[d][3]&255]==1) 
	    if (bitmap3[myhash[d][4]&255][myhash[d][5]&255]==1)
	    if (bitmap4[myhash[d][6]&255][myhash[d][7]&255]==1)
	    while (mylist) 
	    {
		for (a=0;a<vectorsize;a++) 
		{	
		    if (memcmp(mylist->hash, myhash[a], hash_ret_len) == 0)
		    {

			#ifdef DEBUG
			hlog("Cracked uname: %s plaintext: %s thread: %d\n",mylist->username, plain, whoami);
			#endif
			flag = 0;
			pthread_mutex_lock(&crackedmutex);
			addlist = cracked_list;
			while (addlist)
			{
			    if ( (strcmp(addlist->username, mylist->username) == 0) && (memcmp(addlist->hash, mylist->hash, hash_ret_len) == 0)
				&& (strcmp(addlist->salt, mylist->salt) == 0))
				    flag = 1;
			    addlist = addlist->next;
			}
			pthread_mutex_unlock(&crackedmutex);
			if (flag == 0) 
			{
			    add_cracked_list(mylist->username, mylist->hash, mylist->salt, plain[a]);
			    /*plain[num][0]=0;
			    plain[num][1]=0;
			    hash_queue[whoami][lens].lens[num]=2;*/
			}
		    }
		}
		if (mylist) mylist = mylist->next;
	    }
	    for (d=0;d<vectorsize;d++) 
	    {
		hash_queue[whoami][lens].lens[d]--;
	    }
	}
	hash_thread_processed[whoami][0]+=vectorsize*strlen(markov_charset);
    }

}





/* Worker thread entrypoint */
static void * worker_thread_action(void *arg)
{
    int flag = 0;
    unsigned int complex;
    struct  hash_list_s *amylist;
    cpu_set_t set;
    int cnt;
    
    memcpy(&complex,arg,sizeof(int));
    free(worker_thread_keys[complex]);
    amylist = hash_list;

#ifdef HAVE_SSE2
    if (complex == 1) hlog("%s accelerated routines available\n","SSE2");
#endif

    while (amylist)
    {
	if (strlen(amylist->salt) > 0) flag = 1;
	amylist = amylist->next;
    }

    if (flag == 0)
    {
	if (complex == 1) hlog("Attack has linear complexity%s\n","");
	if (attack_method == attack_method_simple_bruteforce) thread_attack_worker_linear_bruteforce(&complex);
	else if (attack_method == attack_method_markov) thread_attack_worker_linear_markov(&complex);
	else thread_attack_worker_linear(&complex);
    }
    else
    {
	if (complex == 1) hlog("Attack has non-linear complexity%s\n","");
	if (attack_method == attack_method_simple_bruteforce) thread_attack_worker_nonlinear_bruteforce(&complex);
	else if (attack_method == attack_method_markov) thread_attack_worker_nonlinear_markov(&complex);
	else thread_attack_worker_nonlinear(&complex);
    }

    pthread_exit(NULL);
    return 0;
}



/* hex2str */
static void hex2str(char *str, char *hex, int len)
{
    int cnt, cnt1;
    unsigned char val=0;
    unsigned char tmp1=0,tmp2=0;
    char *charset="0123456789abcdef";
    bzero(str, (len/2));

    for (cnt=0;cnt<(len/2);cnt++)
    {
        val = 0;
        for (cnt1=0;cnt1<16;cnt1++) if (charset[cnt1] == hex[cnt*2]) tmp1 = cnt1;
        for (cnt1=0;cnt1<16;cnt1++) if (charset[cnt1] == hex[cnt*2+1]) tmp2 = cnt1;
        val |= (tmp1 << 4);
        val |= tmp2;

        *(str+cnt) = val & 255;
    }
}




/* Init hash bitmaps */
static void init_bitmaps()
{
    int a,b;
    char c;
    struct hash_list_s *mylist;
    char hex1[24];
    char hex[8];

    for (a=0;a<255;a++) for (b=0;b<255;b++)
    {
	    bitmap1[a][b]=0;
	    bitmap2[a][b]=0;
	    bitmap3[a][b]=0;
	    bitmap4[a][b]=0;
    }

    mylist = hash_list;
    while (mylist) 
    {
	memcpy(hex1,mylist->hash,8);
	bitmap1[hex1[0]&255][hex1[1]&255]=1;
	bitmap2[hex1[2]&255][hex1[3]&255]=1;
	bitmap3[hex1[4]&255][hex1[5]&255]=1;
	bitmap4[hex1[6]&255][hex1[7]&255]=1;
	if (mylist) mylist = mylist->next;
    }
    hlog("Initialized hash bitmaps\n%s","");
}

/* Spawn worker threads, monitor thread */
hash_stat spawn_threads(unsigned int num, unsigned int queue_size)
{
    unsigned int cnt, cnt1, cnt2;
    int status, status1 = 0;
    pthread_t monitorthread;
    pthread_t monitorinfothread;

    attack_over = 0;
    
    qflag = 0;
    hash_len = hash_plugin_hash_length();
    if (num < 1) return hash_err;
    if (queue_size > HASHKILL_MAXQUEUESIZE) return hash_err;
    if (init_mutexes() == hash_err) return hash_err;
    for (cnt=0;cnt<vectorsize;cnt++) fillbuf[cnt] = malloc(HASHFILE_MAX_PLAIN_LENGTH*5);
    if (hash_plugin_is_special() == 0) init_bitmaps();
    
    for (cnt=0; cnt<num; cnt++)
    {
	for (cnt1=0; cnt1 < queue_size; cnt1++)
	{
	    for (cnt2=0;cnt2<vectorsize;cnt2++) 
	    {
//		hash_queue[cnt][cnt1].plaintext[cnt2] = malloc(HASHFILE_MAX_PLAIN_LENGTH*5);
		posix_memalign(&hash_queue[cnt][cnt1].plaintext[cnt2],16, HASHFILE_MAX_PLAIN_LENGTH*5);
		bzero(hash_queue[cnt][cnt1].plaintext[cnt2],HASHFILE_MAX_PLAIN_LENGTH*5); // make valgrind happy
	    }
	    pushready[cnt][0] = 0;
	    popready[cnt][0] = 0;

	    if (!hash_queue[cnt][cnt1].plaintext)
	    {
		elog("Not enough memory to create thread queues (thread: %d fifo_elem: %d)", cnt, cnt1);
		status1 = 1;
		break;
	    }
	}
	SSL_thread_setup();
	
	hash_thread_queue_counts[cnt][0] = 0;
	hash_thread_processed[cnt][0] = 0;
	hash_thread_cracked[cnt] = 0;
	worker_thread_keys[cnt] = malloc(sizeof(int));
	*(worker_thread_keys[cnt]) = cnt;
	status = pthread_create(&hash_threads[cnt], NULL, worker_thread_action, (void *)worker_thread_keys[cnt]);

	if (status != 0) elog("Cannot spawn thread %d\n",cnt);
	if ((status != 0) || (status1 == 1))
	{
	    for (cnt=0; cnt < num; cnt++)
		for (cnt1=0; cnt1 < queue_size; cnt1++)
		    for (cnt2=0; cnt1 < vectorsize; cnt1++)
			if (hash_queue[cnt][cnt1].plaintext[cnt2] != NULL) free(hash_queue[cnt][cnt1].plaintext[cnt2]);
	    return hash_err;
	}
    }
    hash_num_threads = num;
    hash_queue_size = queue_size;
    cnt++;
    pthread_create(&monitorthread, NULL, start_monitor_thread, &cnt);
    pthread_create(&monitorinfothread, NULL, start_monitor_info_thread, &cnt);
    hlog("Spawned worker threads%s\n", "");
    return hash_ok;
}



/* Get number of CPUs, read from /proc/cpuinfo */
unsigned int hash_num_cpu(void)
{
    FILE *fcpu;
    unsigned int proc = 0;
    char buf[128];
    char *status;

    bzero(buf,128); // Make valgrind happy
    fcpu=fopen("/proc/cpuinfo","r");
    if (!fcpu)
    {
        elog("Cannot open /proc/cpuinfo! errno=%d", errno);
        return 0;
    }
    else
    {
        do
        {
            status = fgets(buf, 128, fcpu);
            if ((status) && (strstr(buf, "processor"))) proc++;
        } while (status);

        fclose(fcpu);
    }
    hlog("Detected %d CPUs.\n", proc);
    return proc;
}



/* Get physical memory in MB */
unsigned long hash_get_total_memory(void)
{
    struct sysinfo info;
    unsigned long memory_mb;

    sysinfo(&info);
    memory_mb = ((info.totalram * info.mem_unit) / (1024*1024));
    hlog("Total physical memory : %ld MB\n", memory_mb);
    return memory_mb;
}



/* Initialize synchronization primitives */
hash_stat init_mutexes(void)
{
    int cnt;
    pthread_mutexattr_t mutexattr;


    for (cnt=0;cnt < hash_num_threads;cnt++)
    {
	pthread_mutexattr_init(&mutexattr); 
	pthread_mutexattr_settype(&mutexattr, PTHREAD_MUTEX_NORMAL); 
	pthread_mutexattr_setpshared(&mutexattr, PTHREAD_PROCESS_SHARED);
	pthread_mutexattr_setprotocol(&mutexattr, PTHREAD_PRIO_NONE);
	
	if (pthread_mutex_init(&queuelock[cnt], &mutexattr) != 0)
	{
    	    elog("Cannot create queue mutex %s\n","");
    	    return hash_err;
	}
    }
    pthread_mutexattr_init(&mutexattr); 
    pthread_mutexattr_settype(&mutexattr, PTHREAD_MUTEX_ADAPTIVE_NP); 
    pthread_mutexattr_setpshared(&mutexattr, PTHREAD_PROCESS_PRIVATE);
    pthread_mutexattr_setprotocol(&mutexattr, PTHREAD_PRIO_NONE);

    if (pthread_mutex_init(&listmutex, &mutexattr))
    {
        elog("Cannot create list mutex %s\n","");
        return hash_err;
    }
    if (pthread_mutex_init(&crackedmutex, &mutexattr))
    {
        elog("Cannot create list mutex %s\n", "");
        return hash_err;
    }
    hlog("Initialized thread mutexes%s\n", "");
    return hash_ok;
}



/* Start monitor thread */ 
static void * start_monitor_thread(void *arg)
{

    sleep(1); 
    printf("\nAttack statistics...\n");
#ifdef DEBUG
    int overall, o1,o2,o3,o4;
    int p1,p2,p3,p4;

    sleep(2);
    while (attack_over == 0)
    {
	overall = 0;
	o1=hash_thread_processed[0][0]/2;
	o2=hash_thread_processed[1][0]/2;
	o3=hash_thread_processed[2][0]/2;
	o4=hash_thread_processed[3][0]/2;
	hash_thread_processed[0][0]=0;
	hash_thread_processed[1][0]=0;
	hash_thread_processed[2][0]=0;
	hash_thread_processed[3][0]=0;
	p1 = hash_thread_queue_counts[0][0];
	p2 = hash_thread_queue_counts[1][0];
	p3 = hash_thread_queue_counts[2][0];
	p4 = hash_thread_queue_counts[3][0];




	printf("(procsd): %d  %d  %d  %d\n",o1,o2,o3,o4);
	printf("(queues): %d  %d  %d  %d\n",p1,p2,p3,p4);
	sleep(2);

    }
    printf("\n");
    return 0;
#else
    int sum,cracked;
    int cnt;
    struct hash_list_s *mylist;
    FILE *sessionfile;
    char key[2];

    
    qflag = 0;
    while (attack_over == 0)
    {
	sleep(3);
	qflag++;
	if (qflag > 4) qflag = 4;
	sum = 0;
	if (attack_over == 0)
	{
	    /* show stats */
	    for (cnt = 0; cnt < hash_num_threads; cnt++)
	    {
		sum += hash_thread_processed[cnt][0];
		hash_thread_processed[cnt][0] = 0;
	    }
	    cracked = 0;
	    pthread_mutex_lock(&crackedmutex);
	    mylist = cracked_list;
	    while (mylist)
	    {
		mylist = mylist->next;
		cracked++;
	    }
	    
	    pthread_mutex_unlock(&crackedmutex);
	    /* This is stupid I know, should be fixed some day */
	    if (attack_overall_count == 1)
	    {
		if ((sum / 2000) > 5) printf("\rSpeed: %d KPlaintexts/sec   Cracked: %d hashes   \b\b\b", (sum / 3000), cracked);
		else printf("\rSpeed: %d plaintexts/sec   Cracked: %d hashes   ", (sum / 3), cracked);
	    }
	    else if ((sum / 2000) > 5)
	    {
		if ( ((attack_current_count*100) / attack_overall_count) > 100)
		{
		    printf("\rProgress: 100%%   Speed: %d KPlaintexts/sec   Cracked: %d hashes   ", (sum / 3000), cracked);
		}
		else  printf("\rProgress: %lld%%   Speed: %d KPlaintexts/sec   Cracked: %d hashes   ", ((attack_current_count*100)/attack_overall_count) ,(sum / 3000), cracked);
	    }
	    else
	    {
		if ( ((attack_current_count*100) / attack_overall_count) > 100)
		{
		    printf("\rProgress: 100%%   Speed: %d plaintexts/sec   Cracked: %d hashes   ", (sum / 3), cracked);
		}
		else  printf("\rProgress: %lld%%   Speed: %d plaintexts/sec   Cracked: %d hashes   ", ((attack_current_count*100)/attack_overall_count) ,(sum / 3), cracked);
	    }
	    fflush(stdout);

	    
	    /* save session : TBD: check error status*/
	    if (session_init_file(&sessionfile) == hash_ok)
	    {
		session_write_parameters(get_current_plugin(), attack_method, ((attack_current_count*100)/attack_overall_count), sessionfile);
		if (attack_method == attack_method_dictionary)
		{
		    session_write_dictionary_parm(attack_current_count, hash_dictionary_file, sessionfile);
		}
		else if (attack_method == attack_method_simple_bruteforce)
		{
		    session_write_bruteforce_parm(bruteforce_start, bruteforce_end, bruteforce_prefix, bruteforce_suffix, bruteforce_charset, attack_current_length, attack_current_str, attack_current_count, sessionfile);
		}
		else if (attack_method == attack_method_hybrid)
		{
		    session_write_hybrid_parm(hash_dictionary_file, bruteforce_start, bruteforce_end, bruteforce_charset, attack_current_count, attack_current_str,  sessionfile);
		}
		else if (attack_method == attack_method_markov)
		{
		    session_write_markov_parm(markov_statfile, markov_threshold, markov_max_len, attack_overall_count, attack_current_count, attack_current_str,  sessionfile);
		}


		if (attack_over==0)
		{
		    session_write_hashlist(sessionfile);
		    session_write_crackedlist(sessionfile);
		}
		session_close_file(sessionfile);
	    }
	    if (cracked >= hashes_count) attack_over = 2;
	}

    }
    printf("\n");
    pthread_exit(NULL);
    return 0;
#endif
}


/* Start monitor info thread (this actually just waits for keyboard input and prints out stats*/ 
static void * start_monitor_info_thread(void *arg)
{
    
    char key[2];
    uint64_t timeest;
    int years,months,days,hours,min,sec;

    char *stringest = alloca(200);
    char *stringest1 = alloca(100);
    
    while (1)
    {
	read(0, &key,2);
	print_cracked_list();
	hlog(" -= End list =-%s\n","");
	time2 = time(NULL);
	bzero(stringest,200);
	timeest = (((attack_overall_count-attack_current_count)*(time2-time1))/(attack_current_count));
	if (attack_overall_count == 1)
	{
	    strcpy(stringest, "Time remaining: UNKNOWN");
	}
	else
	{
	    strcpy(stringest, "Time remaining: ");
	    if ((timeest / (60*60*24*30*12))>1)
	    {
		sprintf(stringest1,"%d years ",(timeest / (60*60*24*30*12)));
		strcat(stringest, stringest1);
	    }
	    if ((timeest / (60*60*24*30))>1)
	    {
		sprintf(stringest1,"%d months ",(timeest / (60*60*24*30))%(12));
		strcat(stringest, stringest1);
	    }
	    if ((timeest / (60*60*24))>1)
	    {
		sprintf(stringest1,"%d days ",(timeest / (60*60*24))%(30));
		strcat(stringest, stringest1);
	    }
	    if ((timeest / (60*60))>1)
	    {
		sprintf(stringest1,"%d hours ",(timeest / (60*60))%(24));
		strcat(stringest, stringest1);
	    }
	    if ((timeest / (60))>1)
	    {
		sprintf(stringest1,"%d minutes ",(timeest / (60))%60);
		strcat(stringest, stringest1);
	    }
	    else 
	    {
		sprintf(stringest1,"%d sec ",(timeest));
		strcat(stringest, stringest1);
	    }
	}
	hlog("%s\n\n", stringest);
    }
}



/* Main thread - bruteforce attack - feed workers */
static inline void bruteforce_fillqueue(char *plaintext)
{
    register unsigned int cnt = 0;
    int flag;
    int val;
    int len;
    int cnt1;
    static int cursize=0;
    
    
    if ((attack_over == 0))
    {
        attack_current_str = plaintext;
        if (attack_method != attack_method_hybrid) attack_current_count++;
	strcpy(fillbuf[cursize],plaintext);
	cursize++;
	if (cursize < vectorsize) return;
	cursize = 0;
        do
        {
	    cnt++;
	    flag = 0;
	    if (cnt == hash_num_threads) cnt = 0;

	    if ((pushready[cnt][0] == 1) || ((hash_thread_queue_counts[cnt][0] == 0)))
	    {
		pushqueue(fillbuf,cnt);
		flag = 1;
	    }
	} while (flag == 0);
    }
}





static void bf_permute_1(char *str, const char *baseset, int len)
{
    int bfplen,bfslen,a1;
    bfplen = strlen(bruteforce_prefix);
    bfslen = strlen(bruteforce_suffix);

    for (a1=saved_bf_string[0]; a1 < len; a1++)
    {
	*(str) = baseset[a1];
	*(str+1) = 0;
	//if (bfslen) strcat(str,bruteforce_suffix);
	bruteforce_fillqueue(str-bfplen);
    }
}

static void bf_permute_2(char *str, const char *baseset, int len)
{
    int bfplen,bfslen,a1,a2;
    bfplen = strlen(bruteforce_prefix);
    bfslen = strlen(bruteforce_suffix);

    for (a1=saved_bf_string[0]; a1 < len; a1++)
    for (a2=saved_bf_string[1]; a2 < len; a2++)
    {
	*(str) = baseset[a1];
	*(str+1) = baseset[a2];
	*(str+2) = 0;
	//if (bfslen) strcat(str,bruteforce_suffix);
	bruteforce_fillqueue(str-bfplen);
    }
}

static void bf_permute_3(char *str, const char *baseset, int len)
{
    int bfplen,bfslen,a1,a2,a3;
    bfplen = strlen(bruteforce_prefix);
    bfslen = strlen(bruteforce_suffix);

    for (a1=saved_bf_string[0]; a1 < len; a1++)
    for (a2=saved_bf_string[1]; a2 < len; a2++)
    for (a3=saved_bf_string[2]; a3 < len; a3++)
    {
	if (saved_bf_string[1]!=0) bzero(saved_bf_string,6);
	*(str) = baseset[a1];
	*(str+1) = baseset[a2];
	*(str+2) = baseset[a3];
	*(str+3) = 0;
	//if (bfslen) strcat(str,bruteforce_suffix);
	bruteforce_fillqueue(str-bfplen);
    }
}

static void bf_permute_4(char *str, const char *baseset, int len)
{
    int bfplen,bfslen,a1,a2,a3,a4;
    bfplen = strlen(bruteforce_prefix);
    bfslen = strlen(bruteforce_suffix);

    for (a1=saved_bf_string[0]; a1 < len; a1++)
    for (a2=saved_bf_string[1]; a2 < len; a2++)
    for (a3=saved_bf_string[2]; a3 < len; a3++)
    for (a4=saved_bf_string[3]; a4 < len; a4++)

    {
	if (saved_bf_string[1]!=0) bzero(saved_bf_string,6);
	*(str) = baseset[a1];
	*(str+1) = baseset[a2];
	*(str+2) = baseset[a3];
	*(str+3) = baseset[a4];
	*(str+4) = 0;
	//if (bfslen) strcat(str,bruteforce_suffix);
	bruteforce_fillqueue(str-bfplen);
    }
}

static void bf_permute_5(char *str, const char *baseset, int len)
{
    int bfplen, bfslen,a1,a2,a3,a4,a5;
    bfplen = strlen(bruteforce_prefix);
    bfslen = strlen(bruteforce_suffix);

    for (a1=saved_bf_string[0]; a1 < len; a1++)
    for (a2=saved_bf_string[1]; a2 < len; a2++)
    for (a3=saved_bf_string[2]; a3 < len; a3++)
    for (a4=saved_bf_string[3]; a4 < len; a4++)
    for (a5=saved_bf_string[4]; a5 < len; a5++)

    {
	*(str) = baseset[a1];
	*(str+1) = baseset[a2];
	*(str+2) = baseset[a3];
	*(str+3) = baseset[a4];
	*(str+4) = baseset[a5];
	*(str+5) = 0;
	//if (bfslen) strcat(str,bruteforce_suffix);
	bruteforce_fillqueue(str-bfplen);
	if (saved_bf_string[1]!=0) bzero(saved_bf_string,6);
    }
    

}

/* this is REALLY getting ugly... */
static void bf_permute_6(char *str, const char *baseset, int len)
{
    int bfplen,bfslen,a1,a2,a3,a4,a5,a6;
    bfplen = strlen(bruteforce_prefix);
    bfslen = strlen(bruteforce_suffix);
    
    for (a1=saved_bf_string[0]; a1 < len; a1++)
    {
    *(str) = *(baseset+a1);
    for (a2=saved_bf_string[1]; a2 < len; a2++)
    {
    *(str+1) = *(baseset+a2);
    for (a3=saved_bf_string[2]; a3 < len; a3++)
    {
    *(str+2) = *(baseset+a3);
    for (a4=saved_bf_string[3]; a4 < len; a4++)
    {
    *(str+3) = *(baseset+a4);
    for (a5=saved_bf_string[4]; a5 < len; a5++)
    {
    *(str+4) = *(baseset+a5);
    for (a6=saved_bf_string[5]; a6 < len; a6++)
    {
	if (saved_bf_string[1]!=0) bzero(saved_bf_string,6);
	*(str+5) = *(baseset+a6);
	*(str+6) = 0;
	bruteforce_fillqueue(str-bfplen);
    }}}}}}
}


static void bf_permute_7(char *str, const char *baseset, int len)
{
    int bfplen,bfslen,a1,a2,a3,a4,a5,a6,a7;
    bfplen = strlen(bruteforce_prefix);
    bfslen = strlen(bruteforce_suffix);
    
    for (a1=saved_bf_string[0]; a1 < len; a1++)
    {
    *(str) = *(baseset+a1);
    for (a2=saved_bf_string[1]; a2 < len; a2++)
    {
    *(str+1) = *(baseset+a2);
    for (a3=saved_bf_string[2]; a3 < len; a3++)
    {
    *(str+2) = *(baseset+a3);
    for (a4=saved_bf_string[3]; a4 < len; a4++)
    {
    *(str+3) = *(baseset+a4);
    for (a5=saved_bf_string[4]; a5 < len; a5++)
    {
    *(str+4) = *(baseset+a5);
    for (a6=saved_bf_string[5]; a6 < len; a6++)
    {
    *(str+5) = *(baseset+a6);
    for (a7=saved_bf_string[6]; a7 < len; a7++)
    {
	if (saved_bf_string[1]!=0) bzero(saved_bf_string,16);
	*(str+6) = *(baseset+a7);
	*(str+7) = 0;
	bruteforce_fillqueue(str-bfplen);
    }}}}}}}
}


static void bf_permute_8(char *str, const char *baseset, int len)
{
    int bfplen,bfslen,a1,a2,a3,a4,a5,a6,a7,a8;
    bfplen = strlen(bruteforce_prefix);
    bfslen = strlen(bruteforce_suffix);
    
    for (a1=saved_bf_string[0]; a1 < len; a1++)
    {
    *(str) = *(baseset+a1);
    for (a2=saved_bf_string[1]; a2 < len; a2++)
    {
    *(str+1) = *(baseset+a2);
    for (a3=saved_bf_string[2]; a3 < len; a3++)
    {
    *(str+2) = *(baseset+a3);
    for (a4=saved_bf_string[3]; a4 < len; a4++)
    {
    *(str+3) = *(baseset+a4);
    for (a5=saved_bf_string[4]; a5 < len; a5++)
    {
    *(str+4) = *(baseset+a5);
    for (a6=saved_bf_string[5]; a6 < len; a6++)
    {
    *(str+5) = *(baseset+a6);
    for (a7=saved_bf_string[6]; a7 < len; a7++)
    {
    *(str+6) = *(baseset+a7);
    for (a8=saved_bf_string[7]; a8 < len; a8++)
    {
	if (saved_bf_string[1]!=0) bzero(saved_bf_string,16);
	*(str+7) = *(baseset+a8);
	*(str+8) = 0;
	bruteforce_fillqueue(str-bfplen);
    }}}}}}}}
}


static inline void bf_permute_9(char *str, const char *baseset, int len)
{
    int bfplen,bfslen,a1,a2,a3,a4,a5,a6,a7,a8,a9;
    bfplen = strlen(bruteforce_prefix);
    bfslen = strlen(bruteforce_suffix);
    
    for (a1=saved_bf_string[0]; a1 < len; a1++)
    {
    *(str) = *(baseset+a1);
    for (a2=saved_bf_string[1]; a2 < len; a2++)
    {
    *(str+1) = *(baseset+a2);
    for (a3=saved_bf_string[2]; a3 < len; a3++)
    {
    *(str+2) = *(baseset+a3);
    for (a4=saved_bf_string[3]; a4 < len; a4++)
    {
    *(str+3) = *(baseset+a4);
    for (a5=saved_bf_string[4]; a5 < len; a5++)
    {
    *(str+4) = *(baseset+a5);
    for (a6=saved_bf_string[5]; a6 < len; a6++)
    {
    *(str+5) = *(baseset+a6);
    for (a7=saved_bf_string[6]; a7 < len; a7++)
    {
    *(str+6) = *(baseset+a7);
    for (a8=saved_bf_string[7]; a8 < len; a8++)
    {
    *(str+7) = *(baseset+a8);
    for (a9=saved_bf_string[8]; a9 < len; a9++)
    {
	if (saved_bf_string[1]!=0) bzero(saved_bf_string,16);
	*(str+8) = *(baseset+a9);
	*(str+9) = 0;
	//if (bfslen) strcat(str,bruteforce_suffix);
	bruteforce_fillqueue(str-bfplen);
    }}}}}}}}}
}

static void bf_permute_10(char *str, const char *baseset, int len)
{
    int bfplen,bfslen,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10;
    bfplen = strlen(bruteforce_prefix);
    bfslen = strlen(bruteforce_suffix);
    
    for (a1=saved_bf_string[0]; a1 < len; a1++)
    {
    *(str) = *(baseset+a1);
    for (a2=saved_bf_string[1]; a2 < len; a2++)
    {
    *(str+1) = *(baseset+a2);
    for (a3=saved_bf_string[2]; a3 < len; a3++)
    {
    *(str+2) = *(baseset+a3);
    for (a4=saved_bf_string[3]; a4 < len; a4++)
    {
    *(str+3) = *(baseset+a4);
    for (a5=saved_bf_string[4]; a5 < len; a5++)
    {
    *(str+4) = *(baseset+a5);
    for (a6=saved_bf_string[5]; a6 < len; a6++)
    {
    *(str+5) = *(baseset+a6);
    for (a7=saved_bf_string[6]; a7 < len; a7++)
    {
    *(str+6) = *(baseset+a7);
    for (a8=saved_bf_string[7]; a8 < len; a8++)
    {
    *(str+7) = *(baseset+a8);
    for (a9=saved_bf_string[8]; a9 < len; a9++)
    {
    *(str+8) = *(baseset+a9);
    for (a10=saved_bf_string[9]; a10 < len; a10++)
    {
	if (saved_bf_string[1]!=0) bzero(saved_bf_string,16);
	*(str+9) = *(baseset+a10);
	*(str+10) = 0;
	bruteforce_fillqueue(str-bfplen);
    }}}}}}}}}}
}


static void bf_permute_11(char *str, const char *baseset, int len)
{
    int bfplen,bfslen,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11;
    bfplen = strlen(bruteforce_prefix);
    bfslen = strlen(bruteforce_suffix);
    
    for (a1=saved_bf_string[0]; a1 < len; a1++)
    {
    *(str) = *(baseset+a1);
    for (a2=saved_bf_string[1]; a2 < len; a2++)
    {
    *(str+1) = *(baseset+a2);
    for (a3=saved_bf_string[2]; a3 < len; a3++)
    {
    *(str+2) = *(baseset+a3);
    for (a4=saved_bf_string[3]; a4 < len; a4++)
    {
    *(str+3) = *(baseset+a4);
    for (a5=saved_bf_string[4]; a5 < len; a5++)
    {
    *(str+4) = *(baseset+a5);
    for (a6=saved_bf_string[5]; a6 < len; a6++)
    {
    *(str+5) = *(baseset+a6);
    for (a7=saved_bf_string[6]; a7 < len; a7++)
    {
    *(str+6) = *(baseset+a7);
    for (a8=saved_bf_string[7]; a8 < len; a8++)
    {
    *(str+7) = *(baseset+a8);
    for (a9=saved_bf_string[8]; a9 < len; a9++)
    {
    *(str+8) = *(baseset+a9);
    for (a10=saved_bf_string[9]; a10 < len; a10++)
    {
    *(str+9) = *(baseset+a10);
    for (a11=saved_bf_string[10]; a11 < len; a11++)
    {
	if (saved_bf_string[1]!=0) bzero(saved_bf_string,16);
	*(str+10) = *(baseset+a11);
	*(str+11) = 0;
	bruteforce_fillqueue(str-bfplen);
    }}}}}}}}}}}
}


static void bf_permute_12(char *str, const char *baseset, int len)
{
    int bfplen,bfslen,a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12;
    bfplen = strlen(bruteforce_prefix);
    bfslen = strlen(bruteforce_suffix);
    
    for (a1=saved_bf_string[0]; a1 < len; a1++)
    {
    *(str) = *(baseset+a1);
    for (a2=saved_bf_string[1]; a2 < len; a2++)
    {
    *(str+1) = *(baseset+a2);
    for (a3=saved_bf_string[2]; a3 < len; a3++)
    {
    *(str+2) = *(baseset+a3);
    for (a4=saved_bf_string[3]; a4 < len; a4++)
    {
    *(str+3) = *(baseset+a4);
    for (a5=saved_bf_string[4]; a5 < len; a5++)
    {
    *(str+4) = *(baseset+a5);
    for (a6=saved_bf_string[5]; a6 < len; a6++)
    {
    *(str+5) = *(baseset+a6);
    for (a7=saved_bf_string[6]; a7 < len; a7++)
    {
    *(str+6) = *(baseset+a7);
    for (a8=saved_bf_string[7]; a8 < len; a8++)
    {
    *(str+7) = *(baseset+a8);
    for (a9=saved_bf_string[8]; a9 < len; a9++)
    {
    *(str+8) = *(baseset+a9);
    for (a10=saved_bf_string[9]; a10 < len; a10++)
    {
    *(str+9) = *(baseset+a10);
    for (a11=saved_bf_string[10]; a11 < len; a11++)
    {
    *(str+10) = *(baseset+a11);
    for (a12=saved_bf_string[11]; a12 < len; a12++)
    {
	if (saved_bf_string[1]!=0) bzero(saved_bf_string,16);
	*(str+11) = *(baseset+a12);
	*(str+12) = 0;
	bruteforce_fillqueue(str-bfplen);
    }}}}}}}}}}}}
}




static void bruteforce_permute(char *str, int n, const char *baseset)
{
    switch (n)
    {
	case 1: bf_permute_1(str, baseset, strlen(baseset)); break;
	case 2: bf_permute_2(str, baseset, strlen(baseset)); break;
	case 3: bf_permute_3(str, baseset, strlen(baseset)); break;
	case 4: bf_permute_4(str, baseset, strlen(baseset)); break;
	case 5: bf_permute_5(str, baseset, strlen(baseset)); break;
	case 6: bf_permute_6(str, baseset, strlen(baseset)); break;
	case 7: bf_permute_7(str, baseset, strlen(baseset)); break;
	case 8: bf_permute_8(str, baseset, strlen(baseset)); break;
	case 9: bf_permute_9(str, baseset, strlen(baseset)); break;
	case 10: bf_permute_10(str, baseset, strlen(baseset)); break;
	case 11: bf_permute_11(str, baseset, strlen(baseset)); break;
	case 12: bf_permute_12(str, baseset, strlen(baseset)); break;

	default: break;
    }
}
	


/* Main thread - hybrid attack */
hash_stat main_thread_hybrid(void)
{
    FILE *dictfile;
    char *buf;
    int cnt;
    int flag,val,len;
    struct stat filestat;

    attack_over = 0;
    time1 = time(NULL);
    hlog("Launching hybrid attack\n%s","");

    dictfile = fopen(hash_dictionary_file, "r");
    if (!dictfile)
    {
	elog("Cannot open dictionary file \'%s\'\n", hash_dictionary_file);
	return hash_err;
    }
    
    stat(hash_dictionary_file, &filestat);
    attack_overall_count = filestat.st_size;
    attack_current_count = 0;
    
    /* we are loading a session file? */
    if (session_restore_flag == 1)
    {
	if (fseek(dictfile, dictionary_current_pos, SEEK_SET) != 0)
	{
	    elog("Cannot seek to position %ld in dictionary file!\n", dictionary_current_pos);
	}
	attack_current_count = dictionary_current_pos;
    }
    
    buf = malloc(HASHFILE_MAX_PLAIN_LENGTH+1);
    cnt = 0;
    flag = 0;

    do
    {
	fgets_unlocked(buf,HASHFILE_MAX_LINE_LENGTH,dictfile);
	len = strlen(buf);
	val=0;
	if (buf)
	{

	    attack_current_count += len;
	    bzero(buf+len-1, 8);
	    strcpy(bruteforce_prefix, buf);
	    do
	    {
		flag = 0;
		cnt++;
		if (cnt == hash_num_threads) cnt = 0;

		if ((pushready[cnt][0] == 1) || ((pushready[cnt][0] == 0) && (hash_thread_queue_counts[cnt][0] == 0)))
		{
		    strcpy(fillbuf[0],buf);
		    pushqueue(fillbuf, cnt);
		    flag = 1;
		}

	    } while (flag == 0);

	    /* next try all combinations */
	    if (bruteforce_start>=bruteforce_end) bruteforce_end = bruteforce_start+1;
	    for (len = bruteforce_start; len < bruteforce_end; len++)
	    {
		if (attack_over > 0) break;
		bruteforce_permute((buf+strlen(buf)), len, bruteforce_charset);
	    }
	    
	}
    } while ((!feof(dictfile))&&(hash_list)&&(attack_over == 0));

    // Some slow plugins may be interrupted even though we wait for empty queues!
    sleep(1);

    for (cnt=0; cnt < hash_num_threads; cnt++) pthread_mutex_unlock(&queuelock[cnt]);
    free(buf);
    thread_wait_empty_queues();
    time2 = time(NULL);
    printf("\n");
    hlog("Hybrid attack complete. It took %ld seconds...\n",time2-time1);
    return hash_ok;
}

/* Main thread - bruteforce attack function */
hash_stat main_thread_bruteforce(void)
{
    char *buf;
    int len;
    int cnt;
    int cnt1;

    attack_over = 0;
    time1 = time(NULL);
    bruteforce_prefixlen = strlen(bruteforce_prefix);
    hlog("Launching bruteforce attack\n%s","");
    /* we are not restoring a session? */
    if (session_restore_flag == 0)
    {
	attack_overall_count = 1;
	bzero(saved_bf_string,16);
	for (cnt = bruteforce_start; cnt < bruteforce_end; cnt++)
	{
	    attack_current_length = cnt;
	    attack_overall_count += (powl(strlen(bruteforce_charset),cnt));
	}
	attack_current_count = 1;
	buf = malloc(HASHFILE_MAX_PLAIN_LENGTH+1);
	bzero(buf, HASHFILE_MAX_PLAIN_LENGTH+1); // make valgrind happy
	if (bruteforce_start>=bruteforce_end) bruteforce_end = bruteforce_start+1;
	for (len = bruteforce_start; len < bruteforce_end; len++)
	{
	    if (attack_over > 0) goto out;
	    strcpy(buf,bruteforce_prefix);
	    bruteforce_permute((buf+strlen(bruteforce_prefix)), len, bruteforce_charset);
	}
    }
    /* we are restoring a session then */
    else
    {
	attack_overall_count = 1;
	attack_over = 0;
	for (cnt = bruteforce_start; cnt < bruteforce_end; cnt++)
	{
	    attack_current_length = cnt;
	    attack_overall_count += powl(strlen(bruteforce_charset),cnt);
	}
	attack_current_count = bruteforce_current_position;
	buf = malloc(HASHFILE_MAX_PLAIN_LENGTH+1);
	bzero(buf, HASHFILE_MAX_PLAIN_LENGTH+1); // make valgrind happy
	strcpy(buf, bruteforce_prefix);

	bzero(saved_bf_string,16);
	strcpy(saved_bf_string, attack_current_str);
	/* some boring tricks */
	for (cnt=0;cnt < strlen(saved_bf_string);cnt++)
	{
	    again:
	    for (cnt1=0; cnt1 < strlen(bruteforce_charset);cnt1++)
	        if (saved_bf_string[cnt] == bruteforce_charset[cnt1]) 
	        {
	    	    saved_bf_string[cnt] = cnt1;
		    cnt++;goto again;
		}
	}
	saved_bf_string[0]=0;
	if (bruteforce_start>=bruteforce_end) bruteforce_end = bruteforce_start+1;
	for (len = bruteforce_current_length; len < bruteforce_end; len++)
	{
	    if (attack_over > 0) goto out;
	    strcpy(buf,bruteforce_prefix);
	    bruteforce_permute((buf+strlen(bruteforce_prefix)), len, bruteforce_charset);
	}
    }
    
    out:
    // Some slow plugins may be interrupted even though we wait for empty queues!
    sleep(1);

    for (cnt=0; cnt<hash_num_threads; cnt++) pthread_mutex_unlock(&queuelock[cnt]);
    free(buf);
    thread_wait_empty_queues();
    time2 = time(NULL);
    printf("\n");
    hlog("Bruteforce attack complete. It took %ld seconds...\n",time2-time1);
    return hash_ok;
}



static void markov_fillqueue(char *plaintext)
{
    unsigned int cnt = 0;
    int flag;
    int val;
    int len;
    static int denom=1;
    static char buf[VECTORSIZE][HASHFILE_MAX_PLAIN_LENGTH];
    static int cursize=0;
    


    if ((attack_over == 0))
    {
        attack_current_str = plaintext;
        denom++;
        if (denom==10000)
        {
    	    attack_current_count++;
    	    denom = 1;
    	}
    	strcpy(fillbuf[cursize],plaintext);
	cursize++;
	if (cursize < vectorsize-1) return;
        cursize = 0;

        do
        {
	    cnt++;
	    flag = 0;
	    if (cnt == hash_num_threads) cnt = 0;

	    if ((pushready[cnt][0] == 1) || ((hash_thread_queue_counts[cnt][0] == 0)))
	    {
		pushqueue(fillbuf,cnt);
		flag = 1;
	    }
	} while (flag == 0);
    }
}

static void markov_permute2(void)
{
    int c1,c2;
    char plaintext[16];
    
    for (c1=0;c1<88;c1++) if (markov0[c1]>markov_threshold)
    for (c2=0;c2<88;c2++) if (markov1[c1][c2]>markov_threshold)
    {
	sprintf(plaintext,"%c%c",markov_charset[c1],markov_charset[c2]);
	markov_fillqueue(plaintext);
	if (attack_over > 0 ) return;
    }
}

static void markov_permute3(void)
{
    int c1,c2,c3;
    char plaintext[16];
    
    for (c1=0;c1<88;c1++) if (markov0[c1]>markov_threshold)
    for (c2=0;c2<88;c2++) if (markov1[c1][c2]>markov_threshold)
    for (c3=0;c3<88;c3++) if (markov1[c2][c3]>markov_threshold)
    {
	plaintext[0] = markov_charset[c1];
	plaintext[1] = markov_charset[c2];
	plaintext[2] = markov_charset[c3];
	plaintext[3] = 0;
	markov_fillqueue(plaintext);
	if (attack_over > 0 ) return;
    }
}


static void markov_permute4(void)
{
    int c1,c2,c3,c4;
    char plaintext[16];
    
    for (c1=0;c1<88;c1++) if (markov0[c1]>markov_threshold)
    for (c2=0;c2<88;c2++) if (markov1[c1][c2]>markov_threshold)
    for (c3=0;c3<88;c3++) if (markov1[c2][c3]>markov_threshold)
    for (c4=0;c4<88;c4++) if (markov1[c3][c4]>markov_threshold)
    {
	plaintext[0] = markov_charset[c1];
	plaintext[1] = markov_charset[c2];
	plaintext[2] = markov_charset[c3];
	plaintext[3] = markov_charset[c4];
	plaintext[4] = 0;
	markov_fillqueue(plaintext);
	if (attack_over > 0 ) return;
    }
}


static inline void markov_permute5(void)
{
    unsigned int c1,c2,c3,c4,c5;
    char plaintext[16];
    
    for (c1=0;c1<88;c1++) if (markov0[c1]>markov_threshold)
    for (c2=0;c2<88;c2++) if (markov1[c1][c2]>markov_threshold)
    for (c3=0;c3<88;c3++) if (markov1[c2][c3]>markov_threshold)
    for (c4=0;c4<88;c4++) if (markov1[c3][c4]>markov_threshold)
    for (c5=0;c5<88;c5++) if (markov1[c4][c5]>markov_threshold)
    {
	plaintext[0] = markov_charset[c1];
	plaintext[1] = markov_charset[c2];
	plaintext[2] = markov_charset[c3];
	plaintext[3] = markov_charset[c4];
	plaintext[4] = markov_charset[c5];
	plaintext[5] = 0;
	markov_fillqueue(plaintext);
	if (attack_over > 0 ) return;
    }
}

static inline void markov_permute6(void)
{
    unsigned int c1,c2,c3,c4,c5,c6;
    char plaintext[16];
    
    for (c1=0;c1<88;c1++) if (markov0[c1]>markov_threshold)
    for (c2=0;c2<88;c2++) if (markov1[c1][c2]>markov_threshold)
    for (c3=0;c3<88;c3++) if (markov1[c2][c3]>markov_threshold)
    for (c4=0;c4<88;c4++) if (markov1[c3][c4]>markov_threshold)
    for (c5=0;c5<88;c5++) if (markov1[c4][c5]>markov_threshold)
    for (c6=0;c6<88;c6++) if (markov1[c5][c6]>markov_threshold)
    {
	plaintext[0] = markov_charset[c1];
	plaintext[1] = markov_charset[c2];
	plaintext[2] = markov_charset[c3];
	plaintext[3] = markov_charset[c4];
	plaintext[4] = markov_charset[c5];
	plaintext[5] = markov_charset[c6];
	plaintext[6] = 0;
	markov_fillqueue(plaintext);
	if (attack_over > 0 ) return;
    }
}

static inline void markov_permute7(void)
{
    unsigned int c1,c2,c3,c4,c5,c6,c7;
    char plaintext[16];
    
    for (c1=0;c1<88;c1++) if (markov0[c1]>markov_threshold)
    for (c2=0;c2<88;c2++) if (markov1[c1][c2]>markov_threshold)
    for (c3=0;c3<88;c3++) if (markov1[c2][c3]>markov_threshold)
    for (c4=0;c4<88;c4++) if (markov1[c3][c4]>markov_threshold)
    for (c5=0;c5<88;c5++) if (markov1[c4][c5]>markov_threshold)
    for (c6=0;c6<88;c6++) if (markov1[c5][c6]>markov_threshold)
    for (c7=0;c7<88;c7++) if (markov1[c6][c7]>markov_threshold)
    {
	plaintext[0] = markov_charset[c1];
	plaintext[1] = markov_charset[c2];
	plaintext[2] = markov_charset[c3];
	plaintext[3] = markov_charset[c4];
	plaintext[4] = markov_charset[c5];
	plaintext[5] = markov_charset[c6];
	plaintext[6] = markov_charset[c7];
	plaintext[7] = 0;
	markov_fillqueue(plaintext);
	if (attack_over > 0 ) return;
    }
}


static inline void markov_permute8(void)
{
    unsigned int c1,c2,c3,c4,c5,c6,c7,c8;
    char plaintext[16];
    
    for (c1=0;c1<88;c1++) if (markov0[c1]>markov_threshold)
    for (c2=0;c2<88;c2++) if (markov1[c1][c2]>markov_threshold)
    for (c3=0;c3<88;c3++) if (markov1[c2][c3]>markov_threshold)
    for (c4=0;c4<88;c4++) if (markov1[c3][c4]>markov_threshold)
    for (c5=0;c5<88;c5++) if (markov1[c4][c5]>markov_threshold)
    for (c6=0;c6<88;c6++) if (markov1[c5][c6]>markov_threshold)
    for (c7=0;c7<88;c7++) if (markov1[c6][c7]>markov_threshold)
    for (c8=0;c8<88;c8++) if (markov1[c7][c8]>markov_threshold)
    {
	plaintext[0] = markov_charset[c1];
	plaintext[1] = markov_charset[c2];
	plaintext[2] = markov_charset[c3];
	plaintext[3] = markov_charset[c4];
	plaintext[4] = markov_charset[c5];
	plaintext[5] = markov_charset[c6];
	plaintext[6] = markov_charset[c7];
	plaintext[7] = markov_charset[c8];
	plaintext[8] = 0;
	markov_fillqueue(plaintext);
	if (attack_over > 0 ) return;
    }
}


static inline void markov_permute9(void)
{
    int c1,c2,c3,c4,c5,c6,c7,c8,c9;
    char plaintext[16];
    
    for (c1=0;c1<88;c1++) if (markov0[c1]>markov_threshold)
    for (c2=0;c2<88;c2++) if (markov1[c1][c2]>markov_threshold)
    for (c3=0;c3<88;c3++) if (markov1[c2][c3]>markov_threshold)
    for (c4=0;c4<88;c4++) if (markov1[c3][c4]>markov_threshold)
    for (c5=0;c5<88;c5++) if (markov1[c4][c5]>markov_threshold)
    for (c6=0;c6<88;c6++) if (markov1[c5][c6]>markov_threshold)
    for (c7=0;c7<88;c7++) if (markov1[c6][c7]>markov_threshold)
    for (c8=0;c8<88;c8++) if (markov1[c7][c8]>markov_threshold)
    for (c9=0;c9<88;c9++) if (markov1[c8][c9]>markov_threshold)
    {
	plaintext[0] = markov_charset[c1];
	plaintext[1] = markov_charset[c2];
	plaintext[2] = markov_charset[c3];
	plaintext[3] = markov_charset[c4];
	plaintext[4] = markov_charset[c5];
	plaintext[5] = markov_charset[c6];
	plaintext[6] = markov_charset[c7];
	plaintext[7] = markov_charset[c8];
	plaintext[8] = markov_charset[c9];
	plaintext[9] = 0;
	markov_fillqueue(plaintext);
	if (attack_over > 0 ) return;
    }
}

static inline void markov_permute10(void)
{
    int c1,c2,c3,c4,c5,c6,c7,c8,c9,c10;
    char plaintext[16];
    
    for (c1=0;c1<88;c1++) if (markov0[c1]>markov_threshold)
    for (c2=0;c2<88;c2++) if (markov1[c1][c2]>markov_threshold)
    for (c3=0;c3<88;c3++) if (markov1[c2][c3]>markov_threshold)
    for (c4=0;c4<88;c4++) if (markov1[c3][c4]>markov_threshold)
    for (c5=0;c5<88;c5++) if (markov1[c4][c5]>markov_threshold)
    for (c6=0;c6<88;c6++) if (markov1[c5][c6]>markov_threshold)
    for (c7=0;c7<88;c7++) if (markov1[c6][c7]>markov_threshold)
    for (c8=0;c8<88;c8++) if (markov1[c7][c8]>markov_threshold)
    for (c9=0;c9<88;c9++) if (markov1[c8][c9]>markov_threshold)
    for (c10=0;c10<88;c10++) if (markov1[c9][c10]>markov_threshold)

    {
	plaintext[0] = markov_charset[c1];
	plaintext[1] = markov_charset[c2];
	plaintext[2] = markov_charset[c3];
	plaintext[3] = markov_charset[c4];
	plaintext[4] = markov_charset[c5];
	plaintext[5] = markov_charset[c6];
	plaintext[6] = markov_charset[c7];
	plaintext[7] = markov_charset[c8];
	plaintext[8] = markov_charset[c9];
	plaintext[9] = markov_charset[c10];
	plaintext[10] = 0;
	markov_fillqueue(plaintext);
	if (attack_over > 0 ) return;
    }
}

static void markov_permute11(void)
{
    int c1,c2,c3,c4,c5,c6,c7,c8,c9,c10,c11;
    char plaintext[16];
    
    for (c1=0;c1<88;c1++) if (markov0[c1]>markov_threshold)
    for (c2=0;c2<88;c2++) if (markov1[c1][c2]>markov_threshold)
    for (c3=0;c3<88;c3++) if (markov1[c2][c3]>markov_threshold)
    for (c4=0;c4<88;c4++) if (markov1[c3][c4]>markov_threshold)
    for (c5=0;c5<88;c5++) if (markov1[c4][c5]>markov_threshold)
    for (c6=0;c6<88;c6++) if (markov1[c5][c6]>markov_threshold)
    for (c7=0;c7<88;c7++) if (markov1[c6][c7]>markov_threshold)
    for (c8=0;c8<88;c8++) if (markov1[c7][c8]>markov_threshold)
    for (c9=0;c9<88;c9++) if (markov1[c8][c9]>markov_threshold)
    for (c10=0;c10<88;c10++) if (markov1[c9][c10]>markov_threshold)
    for (c11=0;c11<88;c11++) if (markov1[c10][c11]>markov_threshold)
    {
	plaintext[0] = markov_charset[c1];
	plaintext[1] = markov_charset[c2];
	plaintext[2] = markov_charset[c3];
	plaintext[3] = markov_charset[c4];
	plaintext[4] = markov_charset[c5];
	plaintext[5] = markov_charset[c6];
	plaintext[6] = markov_charset[c7];
	plaintext[7] = markov_charset[c8];
	plaintext[8] = markov_charset[c9];
	plaintext[9] = markov_charset[c10];
	plaintext[10] = markov_charset[c11];
	plaintext[11] = 0;
	markov_fillqueue(plaintext);
	if (attack_over > 0 ) return;
    }
}

static void markov_permute12(void)
{
    int c1,c2,c3,c4,c5,c6,c7,c8,c9,c10,c11,c12;
    char plaintext[16];
    
    for (c1=0;c1<88;c1++) if (markov0[c1]>markov_threshold)
    for (c2=0;c2<88;c2++) if (markov1[c1][c2]>markov_threshold)
    for (c3=0;c3<88;c3++) if (markov1[c2][c3]>markov_threshold)
    for (c4=0;c4<88;c4++) if (markov1[c3][c4]>markov_threshold)
    for (c5=0;c5<88;c5++) if (markov1[c4][c5]>markov_threshold)
    for (c6=0;c6<88;c6++) if (markov1[c5][c6]>markov_threshold)
    for (c7=0;c7<88;c7++) if (markov1[c6][c7]>markov_threshold)
    for (c8=0;c8<88;c8++) if (markov1[c7][c8]>markov_threshold)
    for (c9=0;c9<88;c9++) if (markov1[c8][c9]>markov_threshold)
    for (c10=0;c10<88;c10++) if (markov1[c9][c10]>markov_threshold)
    for (c11=0;c11<88;c11++) if (markov1[c10][c11]>markov_threshold)
    for (c12=0;c12<88;c12++) if (markov1[c11][c12]>markov_threshold)
    {
	plaintext[0] = markov_charset[c1];
	plaintext[1] = markov_charset[c2];
	plaintext[2] = markov_charset[c3];
	plaintext[3] = markov_charset[c4];
	plaintext[4] = markov_charset[c5];
	plaintext[5] = markov_charset[c6];
	plaintext[6] = markov_charset[c7];
	plaintext[7] = markov_charset[c8];
	plaintext[8] = markov_charset[c9];
	plaintext[9] = markov_charset[c10];
	plaintext[10] = markov_charset[c11];
	plaintext[11] = markov_charset[c12];
	plaintext[12] = 0;
	markov_fillqueue(plaintext);
	if (attack_over > 0 ) return;
    }
}



static void markov_permute(int n)
{
    switch (n)
    {
	case 2: markov_permute2();break;
	case 3: markov_permute3();break;
	case 4: markov_permute4();break;
	case 5: markov_permute5();break;
	case 6: markov_permute6();break;
	case 7: markov_permute7();break;
	case 8: markov_permute8();break;
	case 9: markov_permute9();break;
	case 10: markov_permute10();break;
	case 11: markov_permute11();break;
	case 12: markov_permute12();break;
    }
}


static uint64_t markov_calculate_overall(int n)
{
    int a1,a2,a3,a4,a5,a6,a7,a8,a9,a10,a11,a12,a13,a14;
    int denom = 1;
    uint64_t overall = 1;


    switch (n)
    {
	case 2:
	    for (a1=0;a1<88;a1++) if (markov0[a1]>markov_threshold)
	    for (a2=0;a2<88;a2++) if (markov1[a1][a2]>markov_threshold)
	    {denom++;if (denom == 10000) {denom=1; overall++;} }
	break;

	case 3:
	    for (a1=0;a1<88;a1++) if (markov0[a1]>markov_threshold)
	    for (a2=0;a2<88;a2++) if (markov1[a1][a2]>markov_threshold)
	    for (a3=0;a3<88;a3++) if (markov1[a2][a3]>markov_threshold)
	    {denom++; if (denom == 10000) {denom=1; overall++;} }
	break;

	case 4:
	    for (a1=0;a1<88;a1++) if (markov0[a1]>markov_threshold)
	    for (a2=0;a2<88;a2++) if (markov1[a1][a2]>markov_threshold)
	    for (a3=0;a3<88;a3++) if (markov1[a2][a3]>markov_threshold)
	    for (a4=0;a4<88;a4++) if (markov1[a3][a4]>markov_threshold)
	    {denom++; if (denom == 10000) {denom=1; overall++;}} 
	break;

	case 5:
	    for (a1=0;a1<88;a1++) if (markov0[a1]>markov_threshold)
	    for (a2=0;a2<88;a2++) if (markov1[a1][a2]>markov_threshold)
	    for (a3=0;a3<88;a3++) if (markov1[a2][a3]>markov_threshold)
	    for (a4=0;a4<88;a4++) if (markov1[a3][a4]>markov_threshold)
	    for (a5=0;a5<88;a5++) if (markov1[a4][a5]>markov_threshold)
	    {denom++; if (denom == 10000) {denom=1; overall++;} }
	break;

	case 6:
	    for (a1=0;a1<88;a1++) if (markov0[a1]>markov_threshold)
	    for (a2=0;a2<88;a2++) if (markov1[a1][a2]>markov_threshold)
	    for (a3=0;a3<88;a3++) if (markov1[a2][a3]>markov_threshold)
	    for (a4=0;a4<88;a4++) if (markov1[a3][a4]>markov_threshold)
	    for (a5=0;a5<88;a5++) if (markov1[a4][a5]>markov_threshold)
	    for (a6=0;a6<88;a6++) if (markov1[a5][a6]>markov_threshold)
	    {denom++; if (denom == 10000) {denom=1; overall++;} }
	break;

	case 7:
	    for (a1=0;a1<88;a1++) if (markov0[a1]>markov_threshold)
	    for (a2=0;a2<88;a2++) if (markov1[a1][a2]>markov_threshold)
	    for (a3=0;a3<88;a3++) if (markov1[a2][a3]>markov_threshold)
	    for (a4=0;a4<88;a4++) if (markov1[a3][a4]>markov_threshold)
	    for (a5=0;a5<88;a5++) if (markov1[a4][a5]>markov_threshold)
	    for (a6=0;a6<88;a6++) if (markov1[a5][a6]>markov_threshold)
	    for (a7=0;a7<88;a7++) if (markov1[a6][a7]>markov_threshold)
	    {denom++; if (denom == 10000) {denom=1; overall++;} }
	break;

	case 8:
	    for (a1=0;a1<88;a1++) if (markov0[a1]>markov_threshold)
	    for (a2=0;a2<88;a2++) if (markov1[a1][a2]>markov_threshold)
	    for (a3=0;a3<88;a3++) if (markov1[a2][a3]>markov_threshold)
	    for (a4=0;a4<88;a4++) if (markov1[a3][a4]>markov_threshold)
	    for (a5=0;a5<88;a5++) if (markov1[a4][a5]>markov_threshold)
	    for (a6=0;a6<88;a6++) if (markov1[a5][a6]>markov_threshold)
	    for (a7=0;a7<88;a7++) if (markov1[a6][a7]>markov_threshold)
	    for (a8=0;a8<88;a8++) if (markov1[a7][a8]>markov_threshold)
	    {denom++; if (denom == 10000) {denom=1; overall++;} }
	break;

	case 9:
	    for (a1=0;a1<88;a1++) if (markov0[a1]>markov_threshold)
	    for (a2=0;a2<88;a2++) if (markov1[a1][a2]>markov_threshold)
	    for (a3=0;a3<88;a3++) if (markov1[a2][a3]>markov_threshold)
	    for (a4=0;a4<88;a4++) if (markov1[a3][a4]>markov_threshold)
	    for (a5=0;a5<88;a5++) if (markov1[a4][a5]>markov_threshold)
	    for (a6=0;a6<88;a6++) if (markov1[a5][a6]>markov_threshold)
	    for (a7=0;a7<88;a7++) if (markov1[a6][a7]>markov_threshold)
	    for (a8=0;a8<88;a8++) if (markov1[a7][a8]>markov_threshold)
	    for (a9=0;a9<88;a9++) if (markov1[a8][a9]>markov_threshold)
	    {denom++; if (denom == 10000) {denom=1; overall++;} }
	break;

	case 10:
	    for (a1=0;a1<88;a1++) if (markov0[a1]>markov_threshold)
	    for (a2=0;a2<88;a2++) if (markov1[a1][a2]>markov_threshold)
	    for (a3=0;a3<88;a3++) if (markov1[a2][a3]>markov_threshold)
	    for (a4=0;a4<88;a4++) if (markov1[a3][a4]>markov_threshold)
	    for (a5=0;a5<88;a5++) if (markov1[a4][a5]>markov_threshold)
	    for (a6=0;a6<88;a6++) if (markov1[a5][a6]>markov_threshold)
	    for (a7=0;a7<88;a7++) if (markov1[a6][a7]>markov_threshold)
	    for (a8=0;a8<88;a8++) if (markov1[a7][a8]>markov_threshold)
	    for (a9=0;a9<88;a9++) if (markov1[a8][a9]>markov_threshold)
	    for (a10=0;a10<88;a10++) if (markov1[a9][a10]>markov_threshold)
	    {denom++; if (denom == 10000) {denom=1; overall++;} }
	break;

	case 11:
	    for (a1=0;a1<88;a1++) if (markov0[a1]>markov_threshold)
	    for (a2=0;a2<88;a2++) if (markov1[a1][a2]>markov_threshold)
	    for (a3=0;a3<88;a3++) if (markov1[a2][a3]>markov_threshold)
	    for (a4=0;a4<88;a4++) if (markov1[a3][a4]>markov_threshold)
	    for (a5=0;a5<88;a5++) if (markov1[a4][a5]>markov_threshold)
	    for (a6=0;a6<88;a6++) if (markov1[a5][a6]>markov_threshold)
	    for (a7=0;a7<88;a7++) if (markov1[a6][a7]>markov_threshold)
	    for (a8=0;a8<88;a8++) if (markov1[a7][a8]>markov_threshold)
	    for (a9=0;a9<88;a9++) if (markov1[a8][a9]>markov_threshold)
	    for (a10=0;a10<88;a10++) if (markov1[a9][a10]>markov_threshold)
            for (a11=0;a11<88;a11++) if (markov1[a10][a11]>markov_threshold)
	    {denom++; if (denom == 10000) {denom=1; overall++;} }
	break;

	case 12:
	    for (a1=0;a1<88;a1++) if (markov0[a1]>markov_threshold)
	    for (a2=0;a2<88;a2++) if (markov1[a1][a2]>markov_threshold)
	    for (a3=0;a3<88;a3++) if (markov1[a2][a3]>markov_threshold)
	    for (a4=0;a4<88;a4++) if (markov1[a3][a4]>markov_threshold)
	    for (a5=0;a5<88;a5++) if (markov1[a4][a5]>markov_threshold)
	    for (a6=0;a6<88;a6++) if (markov1[a5][a6]>markov_threshold)
	    for (a7=0;a7<88;a7++) if (markov1[a6][a7]>markov_threshold)
	    for (a8=0;a8<88;a8++) if (markov1[a7][a8]>markov_threshold)
	    for (a9=0;a9<88;a9++) if (markov1[a8][a9]>markov_threshold)
	    for (a10=0;a10<88;a10++) if (markov1[a9][a10]>markov_threshold)
            for (a11=0;a11<88;a11++) if (markov1[a10][a11]>markov_threshold)
            for (a12=0;a12<88;a12++) if (markov1[a11][a12]>markov_threshold)
	    {denom++; if (denom == 10000) {denom=1; overall++;} }
	break;

	case 13:
	    for (a1=0;a1<88;a1++) if (markov0[a1]>markov_threshold)
	    for (a2=0;a2<88;a2++) if (markov1[a1][a2]>markov_threshold)
	    for (a3=0;a3<88;a3++) if (markov1[a2][a3]>markov_threshold)
	    for (a4=0;a4<88;a4++) if (markov1[a3][a4]>markov_threshold)
	    for (a5=0;a5<88;a5++) if (markov1[a4][a5]>markov_threshold)
	    for (a6=0;a6<88;a6++) if (markov1[a5][a6]>markov_threshold)
	    for (a7=0;a7<88;a7++) if (markov1[a6][a7]>markov_threshold)
	    for (a8=0;a8<88;a8++) if (markov1[a7][a8]>markov_threshold)
	    for (a9=0;a9<88;a9++) if (markov1[a8][a9]>markov_threshold)
	    for (a10=0;a10<88;a10++) if (markov1[a9][a10]>markov_threshold)
            for (a11=0;a11<88;a11++) if (markov1[a10][a11]>markov_threshold)
            for (a12=0;a12<88;a12++) if (markov1[a11][a12]>markov_threshold)
            for (a13=0;a13<88;a13++) if (markov1[a12][a13]>markov_threshold)
	    {denom++; if (denom == 10000) {denom=1; overall++;} }
	break;

	case 14:
	    for (a1=0;a1<88;a1++) if (markov0[a1]>markov_threshold)
	    for (a2=0;a2<88;a2++) if (markov1[a1][a2]>markov_threshold)
	    for (a3=0;a3<88;a3++) if (markov1[a2][a3]>markov_threshold)
	    for (a4=0;a4<88;a4++) if (markov1[a3][a4]>markov_threshold)
	    for (a5=0;a5<88;a5++) if (markov1[a4][a5]>markov_threshold)
	    for (a6=0;a6<88;a6++) if (markov1[a5][a6]>markov_threshold)
	    for (a7=0;a7<88;a7++) if (markov1[a6][a7]>markov_threshold)
	    for (a8=0;a8<88;a8++) if (markov1[a7][a8]>markov_threshold)
	    for (a9=0;a9<88;a9++) if (markov1[a8][a9]>markov_threshold)
	    for (a10=0;a10<88;a10++) if (markov1[a9][a10]>markov_threshold)
            for (a11=0;a11<88;a11++) if (markov1[a10][a11]>markov_threshold)
            for (a12=0;a12<88;a12++) if (markov1[a11][a12]>markov_threshold)
            for (a13=0;a13<88;a13++) if (markov1[a12][a13]>markov_threshold)
            for (a14=0;a14<88;a14++) if (markov1[a13][a14]>markov_threshold)
	    {denom++; if (denom == 10000) {denom=1; overall++;} }
	break;
    }
    return overall;
}


static void *calculate_markov_thread(void *arg)
{
    int cnt;
    uint64_t overall=0;
    
    for (cnt=2;cnt<markov_max_len;cnt++) 
    {
	overall += markov_calculate_overall(cnt);
    }
    attack_overall_count = overall;
    pthread_exit(NULL);
}

/* Main thread - markov attack */
hash_stat main_thread_markov(void)
{
    char *buf;
    int len;
    int cnt;
    int cnt1;
    int cnt2;
    uint64_t tmp,tmp1,tmp2;
    pthread_t calc_thread;

    attack_over = 0;
    time1 = time(NULL);
    hlog("Launching Markov attack\n%s","");
    /* we are not restoring a session? */
    if (session_restore_flag == 0)
    {
	attack_overall_count = 1;
	bzero(saved_bf_string,16);
	hlog("Markov max len: %d threshold:%d\n",markov_max_len, markov_threshold);
	hlog("Progress indicator will be available once Markov calculations are done...\n%s","");
        attack_overall_count = 1;
        pthread_create(&calc_thread, NULL, calculate_markov_thread, NULL);

	for (len = 2; len < markov_max_len; len++)
	{
	    if (attack_over > 0) goto out;
	    markov_permute(len);
	}
    }
    /* we are restoring a session then */
    else
    {
	strcpy(markov_charset,"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789~!{}@#$%^&*()-+[]|\\;':,./");
	markov_load_statfile(markov_statfile);
	attack_over = 0;
	attack_current_count = markov_current_pos;
	buf = malloc(HASHFILE_MAX_PLAIN_LENGTH+1);
	hlog("Markov max len: %d threshold:%d\n",markov_max_len, markov_threshold);
	hlog("Markov current len: %d\n",markov_current_len);

	if (attack_overall_count == 1) 	
	{
	    hlog("Progress indicator will be available once Markov calculations are done...\n%s","");
	    pthread_create(&calc_thread, NULL, calculate_markov_thread, NULL);
	}
	for (len = markov_current_len; len < markov_max_len; len++)
	{
	    if (attack_over > 0) goto out;
	    markov_permute(len);
	}
	/* some boring tricks */
    }
    out:
    // Some slow plugins may be interrupted even though we wait for empty queues!
    sleep(1);
    for (cnt=0; cnt<hash_num_threads; cnt++) pthread_mutex_unlock(&queuelock[cnt]);
    thread_wait_empty_queues();
    time2 = time(NULL);
    printf("\n");
    hlog("Markov attack complete. It took %ld seconds...\n",time2-time1);
    return hash_ok;
}



/* Main thread - dictionary attack function */
hash_stat main_thread_dictionary(void)
{
    FILE *dictfile;
    char *buf;
    int cnt;
    int flag,val,len;
    struct stat filestat;
    int bread, cread, dread;
    char *buf1 = malloc(4096);
    volatile int bufsl=0;

    attack_over = 0;
    time1 = time(NULL);
    hlog("Launching dictionary attack\n%s","");

    dictfile = fopen(hash_dictionary_file, "r");
    if (!dictfile)
    {
	elog("Cannot open dictionary file \'%s\'\n", hash_dictionary_file);
	return hash_err;
    }
    
    stat(hash_dictionary_file, &filestat);
    attack_overall_count = filestat.st_size;
    attack_current_count = 0;
    
    /* we are loading a session file? */
    if (session_restore_flag == 1)
    {
	if (fseek(dictfile, dictionary_current_pos, SEEK_SET)!=0)
	{
	    elog("Cannot seek to position %ld in dictionary file!\n",dictionary_current_pos);
	}
	attack_current_count = dictionary_current_pos;
    }

    buf = malloc(HASHFILE_MAX_PLAIN_LENGTH*5);
    cnt = 0;
    flag = 0;
    cnt = 0;
    do
    {
	bread = fread(buf1, 4096, 1, dictfile);
	cread = 0;dread=0;
	while ( cread < 4096 )
	{
	    *(buf+dread) = *(buf1+cread);
	    cread++;
	    dread++;
	    if ( (dread == HASHFILE_MAX_PLAIN_LENGTH) || (*(buf+dread-1)=='\n'))
	    {
		*(buf+dread-1) = 0;
		//*(buf+dread) = 0;
		dread = 0;
		len = strlen(buf);
		attack_current_count += len;
		if (strlen(buf)>1) strcpy(fillbuf[bufsl],buf);
		bzero(fillbuf[bufsl]+HASHFILE_MAX_PLAIN_LENGTH+1,1);
		bufsl++;
		if (bufsl<vectorsize-1) goto again;
		bufsl=0;

		flag = 0;
		do
		{
		    cnt++;
		    if (cnt == hash_num_threads) cnt = 0;
		    if ((pushready[cnt][0] == 1) || (hash_thread_queue_counts[cnt][0] == 0))
		    {
			pushqueue(fillbuf,cnt);
			flag = 1;
		    }
		} while (flag == 0);
	    }
	    again:
	    dread++;
	    dread--;
	}
	pushqueue(fillbuf,cnt);
	dread = 0;
	len = strlen(buf);
	attack_current_count += len;
	strcpy(fillbuf[bufsl],buf);
	bufsl++;

	flag = 0;
	do
	{
	    cnt++;
	    if (cnt == hash_num_threads) cnt = 0;
	    if ((pushready[cnt][0] == 1) || (hash_thread_queue_counts[cnt][0] == 0))
	    {
		pushqueue(fillbuf,cnt);
		flag = 1;
	    }
	} while (flag == 0);
    } while (bread==1);
    

    for (cnt=0; cnt < hash_num_threads; cnt++) pthread_mutex_unlock(&queuelock[cnt]);
    free(buf);
    free(buf1);
    
    // Some slow plugins may be interrupted even though we wait for empty queues!
    sleep(2);
    
    thread_wait_empty_queues();
    time2 = time(NULL);
    attack_over = 2;

    printf("\n");
    hlog("Dictionary attack complete. It took %ld seconds...\n",time2-time1);
    return hash_ok;
}


void thread_attack_cleanup(void)
{
    int cnt, cnt1, cnt2;
    
/*    for (cnt = 0; cnt <= hash_num_threads; cnt++)
	for (cnt1 = 0; cnt1 < hash_queue_size; cnt1++)
	    for (cnt2 = 0; cnt2 < vectorsize; cnt2++)
		if (hash_queue[cnt][cnt1].plaintext[cnt2])
		    free(hash_queue[cnt][cnt1].plaintext[cnt2]);*/
    attack_over = 2;
}



/* Used by sessions.c to set bf params */
void thread_set_bf_len(int len)
{
    bruteforce_current_length = len;
}



/* Used by sessions.c to set bf params */
void thread_set_bf_str(char *str)
{
    attack_current_str = malloc(strlen(str)+1);
    strcpy(attack_current_str, str);
}



/* Used by sessions.c to set dict params */
void thread_set_dict_pos(long pos)
{
    dictionary_current_pos = pos;
}



/* Used by sessions.c to set bf params */
void thread_set_bf_pos(uint64_t pos)
{
    bruteforce_current_position = pos;
}


/* Used by sessions.c to set markov params */
void thread_set_markov_pos(uint64_t pos)
{
    markov_current_pos = pos;
}


/* Used by sessions.c to set markov params */
void thread_set_markov_max_pos(uint64_t pos)
{
    attack_overall_count = pos;
    
}


/* Used by sessions.c to set markov params */
void thread_set_markov_str(char *str)
{
    attack_current_str = malloc(strlen(str)+1);
    strcpy(attack_current_str, str);
}


/* Used by sessions.c to set markov params */
void thread_set_markov_len(int len)
{
    markov_current_len = len;
}
